program dedup(input, output);
(* dedup: remove character duplications

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of dedup.p 2015 Mar 26
2015 Mar 26, 1.01: handle first character properly
2015 Mar 26, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.dedup *)
(*
name
   dedup: remove character duplications

synopsis
   dedup(input: in, output: out)

files

   input:  input file

   output: input file with duplicated characters removed.

description

   The input file is copied to the output file and when a character
   matches the previous character it is not output.

examples

   The string:
     'hi    therrre'
   becomes:
     'hi there'

documentation

see also

   rembla, rb

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.dedup *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module dedup.themain *)
procedure themain(var fin, fout: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   previous: char; (* a previous character *)
begin
(*
   writeln(output,'dedup ',version:4:2);
*)

   (* Handle first character by making it some character OTHER than
      the first one observed, whatever that is! *)
   if ord(fin^) = 0 then begin
      (* It is safe to add 1 to the nul character since that
         creates the soh character! *)
      previous := chr(ord(fin^)+1);
   end
   else begin
      (* It is safe to subtract 1 to create any other character,
         including the null character! *)
      previous := chr(ord(fin^)-1);
   end;

   while not eof(fin) do begin
      (* copy a line *)
      while (not eoln(fin)) and (not eof(fin)) do begin
         if fin^ <> previous
         then begin
            write(fout, fin^);
            previous := fin^;
         end;
         get(fin)
      end;
      writeln(fout);
      readln(fin)
   end;

end;
(* end module dedup.themain *)

begin
   themain(input, output);
1: end.
