program diffint(input,output);
(* differences between integers.

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

*)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.12; (* of diffint.p 2022 Oct 29
2022 Oct 29, 1.12: report line of non-integers
2022 Oct 29, 1.11: fix documentation, detect non-integer
2022 Oct 29, 1.10: simplify!!!!  No '*' lines.
2022 Aug 24, 1.09: remove the zero '0' at the end of the output!
2022 Aug 24, 1.08: remove the zero '0' at the end of the output!
2007 Nov 29, 1.07: rename difint as diffint
2007 Nov 28, 1.06: rename difint as diffint
1996 Oct 24, 1.05: functional
1986 dec  2, 1.00: origin
*)
(* end module version *)

(* begin module describe.diffint *)
(*
name
   diffint: differences between integers

synopsis
   diffint(input: in, output: out);

files
   input: a set of integers, one per line.

   output: the difference between each integer and the previous one.

describe

   Difference between each integer in input and the
   previous one is given to the output.

documentation

see also

   {Program that allows selection of columns and ranges of differences
    to be selected:} diffintcut.p

author

   Thomas Dana Schneider

bugs

technical notes

   The program originally acted as if the integer before the first one
   is zero, but this creates artifacts.  The initial data point is now
   dropped.  However, data often comes separate into chunks.  This is
   marked by the comments.  So the program resets everytime there is a
   comment line.

   The name was changed to diffint (from difint) because that's
   closer to the Unix diff command.

   The program used to allow:

   "Lines that begin with an asterisk ('*') are first copied to the
   output."
   and
   'Lines that begin with "*" are ignored.'
   This is no longer true.

bugs
   DO NOT GIVE NON-INTEGERS TO THE CODE.
   If you do, it may give you strange differences.
   The code only checks for non-integers in the first character
   of each line.

*)
(* end module describe.diffint *)

   testing = false; (* set to true to test ord function *)

var
   current, previous: integer; (* two integers from the input file *)
   haveprevious: boolean; (* do we have the previous value? *)

   (* variables for checking the input *)
   c: char; (* first character of a line *)
   ordzero, ordnine: integer; (* ord of 0 and 1 *)
   ordc: integer; (* ord of a character from input *)
   line: integer; (* current line number for reporting problems *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

begin

   ordzero := ord('0');
   if testing then writeln(output,'ord(0)=',ordzero:1);
   ordnine := ord('9');
   if testing then writeln(output,'ord(9)=',ordnine:1);

   haveprevious := false;
   line := 0;
   while not eof(input) do begin
      line := line + 1;
      c := input^;
      if not eof(input) then begin
         if (ord(c) < ordzero) or (ord(c) > ordnine) then begin
            writeln(output,'NON DIGIT FOUND IN INPUT:');
            writeln(output,'on line ',line:1,' character = "',c,'"');
            halt;
         end;
      end;
      readln(input, current);
      if not eof(input) then begin
         if haveprevious then begin
            writeln(output,current-previous:10);
         end;
         previous := current;
         haveprevious := true;
      end
   end;

1: end.
