program disum(data, disump, xyin, output);
(* disum: distance vs Ri sum

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.01; (* of disum.p 1996 October 10
origin 1987 Sep 21 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.disum *)
(*
name
   disum: distance vs Ri sum

synopsis
   disum(data: in, disump: in, xyin: out, output: out)

files
   data:  output from scan program

   xyin:  distance between sites found by Scan versus the
      sum of their Ri values.

   disump:  parameters to control the program
      first line: version of disump that this parameter file is designed
         for.

   output: messages to the user

description
   Can one correlate Ri values with distance?

examples

documentation

see also
   scan.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.disum *)

type
   spotptr = ^spot;
   spot = record
      (* a record of the data file *)
      l: integer; (* the position *)
      ri: real; (* the Ri value *)
      next: spotptr;
   end;

var
   data, (* file used by this program *)
   disump, (* file used by this program *)
   xyin: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.57; (@ of prgmod.p 2017 Jan 30 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.57; (@ of prgmod.p 2017 Jan 30 *)

(* begin module disum.themain *)
procedure themain(var data, disump, xyin: text);
(* the main procedure of the program *)
var
   allspots: spotptr; (* all of the data file data *)
   aspot: spotptr; (* one of allspots *)
   bspot: spotptr; (* one of allspots *)
   lastspot: spotptr; (* the last filled entry of allspots *)
   parameterversion: real; (* parameter version number *)
begin
   writeln(output,'disum ',version:4:2);
   reset(disump);
   readln(disump, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   reset(data);

   new(allspots);
   aspot := allspots;
   while not eof(data) do begin
      if data^='*'
      then readln(data)
      else begin
         skipcolumn(data);
         skipcolumn(data);
         skipcolumn(data);
         read(data,aspot^.l);
         skipcolumn(data);
         read(data,aspot^.ri);
         readln(data);
(*
         writeln(output,aspot^.l:1,' ',aspot^.ri:4:1,' bits');
*)
         lastspot := aspot;
         new(aspot^.next);
         aspot := aspot^.next;
      end
   end;
   (* remove the last one and clean up the end *)
   dispose(lastspot^.next);
   lastspot^.next := nil;

(* Show the entire list
   aspot := allspots;
   while aspot <> nil do begin
      writeln(output,aspot^.l:1,' ',aspot^.ri:4:1,' bits');
      aspot := aspot^.next;
   end;
*)

   rewrite(xyin);
   writeln(xyin,'* disum ',version:4:2);
   aspot := allspots;
   while aspot <> nil do begin
      writeln(output,aspot^.l:1,' ',aspot^.ri:4:1,' bits');
      bspot := allspots;
      while bspot <> aspot do begin
         writeln(xyin,
                aspot^.l - bspot^.l:10,
            ' ',bspot^.ri + aspot^.ri:10:6);
         bspot := bspot^.next;
      end;

      aspot := aspot^.next;
   end;

end;
(* end module disum.themain *)

begin
   themain(data, disump, xyin);
1: end.



























