program loocat(cat, list, output);
(* look at a catalogue
      by gary stormo
      modified by thomas schneider
      copyright 1985
      module library: delman
*)

const
(* begin module version *)
   version = 1.18; (* of loocat.p 2017 Jul 14
2017 Jul 14, 1.18: namelength/[linelength not here] set to 200/210 (see catal)
2017 Jul 10, 1.17: namelength set to 200 to allow even longer names
2001 Mar 16, 1.16: only show names to their length
1999 May  4, 1.15: reverse order so listing list is cleaner
origin before 1982 august 14 *)
(* end module version *)

(* begin module describe.loocat *)
(*
name
      loocat: look at a catalogue

synopsis
      loocat(cat: in, list: out, output: out)

files
      cat: a catalogue generated by the catal program
      list: a listing of the contents of cat
      output: messages to the user

description
      Loocat allows one to look at a catalogue that the librarian
      delila normally looks at.  These catalogues are files of a special
      type of record (called item) so that delila can read the information
      rapidly.  However this makes it difficult to see what the catalogue
      contains.  Loocat is useful for understanding or debugging catalogues.

documentation
      libdef, delman.construction.catal

see also
      catal.p, delila.p

author
      Gary Stormo and Thomas Schneider

bugs
      none known

*)
(* end module describe.loocat *)

      namelength = 200; (* the maximum length of names in a library *)

type
      alpha = packed array[1..namelength] of char; (* a set of letters *)

      name = record (* a name in a library *)
         letters: alpha;
         length: 0..namelength
      end;

      item = record (* information about an object in a library *)
         nodeletter: char; (* identification of library object type *)
         nam: name;  (* the name of the object *)
         line: integer (* the place of the object in the library *)
      end;

      recordfile = file of item;

var
      cat: recordfile; (* a catalogue in the delila system *)
      list: text; (* list of items in cat *)
      x: item; (* one item in cat *)

begin (* loocat *)
      writeln(output, ' loocat ', version:4:2);

      rewrite(list);
      writeln(list, ' loocat ', version:4:2,' look at a catalogue');
      writeln(list);
      reset(cat);

      if not eof(cat) then begin
         writeln(list,
                      ' length',
                      '  line',
                      ' id',
                      ' name',' ':(namelength-5)
                      );

         while not eof(cat) do
         begin
            x := cat^; (* pickup the item *)
            get(cat);  (* move along the file *)
            with x do begin
               write  (list, ' ', nam.length:5);
               write  (list, ' ', line:5);
               write  (list, ' ', nodeletter, ' ');
               if nam.length > namelength then begin
                  writeln(output,'warning: name length (',
                         nam.length,') is too big');
                  write  (list, ' ', nam.letters:namelength);
               end
               else begin
                  write  (list, ' ', nam.letters:nam.length);
               end;
               writeln(list)
            end
         end
      end
      else writeln(list,' empty file');
end. (* loocat *)
