program memory(input, memoryp, pienumbers, piechartp, output);
(* memory: analysis of computer memory

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.07; (* of memory.p 2010 Dec 05
2010 Dec 05, 1.07: adjust colors
2010 Dec 04, 1.06: prepare for pie display
2010 Dec 04, 1.05: compute all fractions
2010 Dec 04, 1.04: used is not part of the sum!
2010 Dec 04, 1.03: clean up
2010 Dec 04, 1.02: parse line from top directly to handle M vs G sizes
2010 Dec 04, 1.01: remove use of memoryp
2010 Dec 04, 1.00: origin (functional)
*)
updateversion = 1.07; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.memory *)
(*
name
   memory: analysis of computer memory

synopsis
   memory(input: in, memoryp: in,
          pienumbers: out, piechartp: out, output: out)

files

   input:  line of output from top:

   memoryp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      title of the piechart.  This can be the computer name, uname -n.
      One line.

      font size for the title of the piechart.

   piechart: input to piechart

   pienumbers:  input to piechart

description

   Use the output of top to show computer memory fractions like
   Mac OS X Activity Monitor's pie chart.

examples

documentation


  This program shows the memory in OS X (and maybe other Unix OS)
  using the top program.

  man top:
  
         -l <samples>
                Use logging mode and display <samples> samples, even
                if standard output  is  a  terminal.  0 is treated as
                infinity.  Rather than redisplaying, output is
                periodically printed in raw form.   Note that  the 
                first sample displayed will have an invalid %CPU dis-
                played for each process, as it is calculated using the
                delta between samples.

so

top -l 1 | grep 'PhysMem'

gives

PhysMem: 574M wired, 576M active, 304M inactive, 1454M used, 15G free.

see also

   {mem script}

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.memory *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

var
   {input, (* file used by this program *)}
   memoryp, (* file used by this program *)
   piechartp, (* file used by this program *)
   pienumbers: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

(* holdbegin module string.type *)
(* holdend module string.type *)

(* holdbegin module interact.getstring *)
(* holdend module interact.getstring *)

(* holdbegin module interact.writestring *)
(* holdend module interact.writestring *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.44; (@ of prgmod.p 2010 Dec 04 *)

(* begin module memory.themain *)
procedure themain(var input, memoryp, pienumbers, piechartp: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   title: text; (* internal title for the chart *)
   fontsize: integer; (* font size for the title *)

   wired: real; (* a component of Physical Memory *)
   active: real; (* a component of Physical Memory *)
   inactive: real; (* a component of Physical Memory *)
   used: real; (* a component of Physical Memory *)
   free: real; (* a component of Physical Memory *)
   
   wid: integer; (* width of output numbers *)
   dec: integer; (* decimals of output numbers *)
   sum: real; (* sum of memory parts *)
   freefraction: real; (* fraction of free of memory parts *)
   wiredfraction: real; (* fraction of wired of memory parts *)
   activefraction: real; (* fraction of active of memory parts *)
   inactivefraction: real; (* fraction of inactive of memory parts *)
   MperG: integer; (* Megabytes per Gigabyte *)

   testinput: boolean; (* test the input stream *)
   c: char; (* an input character *)

procedure skipword(var fin, fout: text);
(* skip a word in the file, copy to output *)
var
   c: char; (* a character read from fin *)
begin
   while fin^ = ' ' do begin
      read(fin,c);
      write(fout,c);
   end;
   while fin^ <> ' ' do begin
      read(fin,c);
      write(fout,c);
   end;
end;

begin
   writeln(output,'memory ',version:4:2);
   reset(memoryp);
   readln(memoryp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   rewrite(title);
   copyaline(memoryp, title);
   readln(memoryp, fontsize);

   testinput := true;
   testinput := false;
   if testinput then begin
      while not eoln(input) do begin
         read(input, c);
         write(output, c);
      end;
         writeln(output);
      halt;
   end;

   wid := 9;
   dec := 3;

   (* It's not quite clear what is given by top:
       PhysMem     Physical memory usage, broken into wired, active, inactive,
                   used, and free components
      But google
        gigabyte megabyte
      gives 1024 megabytes per gigabytes as standard
   *)
   MperG := 1024;

{
original purified string method:
top -l 1 |\
grep 'PhysMem' |\
tr -dc '[:digit:] ' |\
cat | memory
   read(input,
   wired,
   active,
   inactive,
   used,
   free);
}

(* The line to parse looks like this:
PhysMem: 574M wired, 576M active, 304M inactive, 1454M used, 15G free.
*)

   skipword(input, output); (* PhysMem: *)

   read(input, wired);
   write(output, ' ',round(wired):1);
   read(input, c);
   write(output, c:1);
   skipword(input, output); (* wired, *)
   if c = 'M' then wired := wired/MperG;

   read(input, active);
   write(output, ' ',round(active):1);
   read(input, c);
   write(output, c:1);
   skipword(input, output); (* active, *)
   if c = 'M' then active := active/MperG;

   read(input, inactive);
   write(output, ' ',round(inactive):1);
   read(input, c);
   write(output, c:1);
   skipword(input, output); (* inactive, *)
   if c = 'M' then inactive := inactive/MperG;

   read(input, used);
   write(output, ' ',round(used):1);
   read(input, c);
   write(output, c:1);
{
writeln(output);
writeln(output, '[',c:1, '] that is c!!');
writeln(output,'ord("M")=',ord('M'));
writeln(output,'ord(c)=',ord(c));
   if c = 'M' then begin used := used/MperG; writeln(output,'BUBBBA'); end;
}
   if c = 'M' then used := used/MperG;
   skipword(input, output); (* used, *)

   read(input, free);
   write(output, ' ',round(free):1);
   read(input, c);
   write(output, c:1);
   skipword(input, output); (* free. *)
   if c = 'M' then free := free/MperG;

   sum := wired + active + inactive + free;
   freefraction := free/sum;
   wiredfraction := wired/sum;
   activefraction := active/sum;
   inactivefraction := inactive/sum;

   writeln(output);
   writeln(output,' ','Gigabytes':wid,    ' KIND');
   writeln(output,' ',free:wid:dec,       ' free     GREEN  ');
   writeln(output,' ',wired:wid:dec,      ' wired    RED');
   writeln(output,' ',active:wid:dec,     ' active   YELLOW');
   writeln(output,' ',inactive:wid:dec,   ' inactive BLUE');
   writeln(output,' ',used:wid:dec,       ' used     (not part of sum)');
   writeln(output,' ','------':wid,       '----');
   writeln(output,' ',sum:wid:dec,        ' SUM');
   writeln(output,' ',freefraction:wid:dec, ' free fraction');

   writeln(output);
   writeln(output,' ',(round(    freefraction*100)):wid, '% free percent');
   writeln(output,' ',(round(   wiredfraction*100)):wid, '% wired percent');
   writeln(output,' ',(round(  activefraction*100)):wid, '% active percent');
   writeln(output,' ',(round(inactivefraction*100)):wid, '% inactive percent');

   rewrite(piechartp);
   writeln(piechartp,'1.03  version of piechart',
                     ' that this parameter file is designed for.');
   writeln(piechartp,'2.0   diameter of the chart (cm)');
   writeln(piechartp,'90    angle to start');
   writeln(piechartp,'-1    direction +1 or -1 to draw');
   reset(title);
   copyaline(title,piechartp);
   writeln(piechartp,fontsize:1,'    font size for the title');

{ for testing
wired := 1;
active := 2;
inactive := 3;
free := 4;
}

   rewrite(pienumbers);
   writeln(pienumbers,' ',   wired:wid:dec,' r 1 0 0 % wired');
   writeln(pienumbers,' ',  active:wid:dec,' r 1 1 0 % active');
   writeln(pienumbers,' ',inactive:wid:dec,' r 0 0 1 % inactive');
   writeln(pienumbers,' ',    free:wid:dec,' r 0 1 0 % free');

end;
(* end module memory.themain *)

begin
   themain(input, memoryp, pienumbers, piechartp);
1: end.
