(* ******** WARNING ********
   nocomequal.p is created AUTOMATICALLY from nocom.p
   by the mknocomequal script.  Do not modify this program by hand! *)

program nocomequal(input, output);
(* nocomequal: remove comments from a program

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1;

const
(* begin module version *)
   version = 1.24; (* of nocomequal.p 2016 Jul 15
2016 Jul 15, 1.24: clean up
2016 Jul 15, 1.23: in equallines mode - no extra lines at top of file!
2016 Jul 15, 1.22: spell nibbon consistently
2016 Jul 15, 1.21: implement equallines
2016 Jul 15, 1.20: clean up
2016 Jul 15, 1.19: fillstring bug? no - tc removed comments in filltrigger!
                   tc uses nocomequal to remove comments and it removed the nibbons
                   {} and [**] (with '['=>'(' and ']' => ')' !)
                   SO: tc does not remove comments from nocomequal AND
                   nocomequal handles the nibbons
2016 Jul 14, 1.18: use file names for themain
2012 Nov 14, 1.17: cleanup
2012 Nov 14, 1.16: cleanup, fix subtle case of one semicolon; document
2012 Nov 14, 1.15: backup
2012 Nov 14, 1.14: remove the nibbons (detecting comment nibbons)
2012 Nov 14, 1.13: remove the nibbons (partially functional)
2012 Nov 14, 1.12: handle lister properly. ignore '{' inside quotes!
2000 Dec  4, 1.11: backup
2000 Dec  4, 1.06: upgrade to mention stdin
2000 Nov 16, 1.06: handle {}
1996 Jul 27, 1.05: previous version
origin 1990 May 14 from decom 1988 Dec 14 *)
(* end module version *)

(* begin module describe.nocomequal *)
(*
name

   nocomequal: remove comments from a program

synopsis
   nocomequal(input: stdin; output: out)

files

   input:  a Pascal or Delila program with comments.

   output: the same program with the contents of the comments removed

description

   This program removes comments from a Delila or Pascal source code so that
   one can, for example, compare two outputs of dbinst.  It can also be used
   to remove comments to allow one to count the number of Delila 'get'
   statements.

   Note:  Generally it is a bad idea to permanently remove comments from
   a program!

see also

   {Description of what stdin means: }
   shell.p

   {Programs mentioned in the description above: }
   dbinst.p
   delila.p

   {Other comment manipulation programs: }
   decom.p
   codecomments.p

   {nocomequal removes extra lines in the code.  To keep extra lines, 
   is} mknocomequalequal { which generates the nocomequal.p program.}

author

   Thomas Dana Schneider

bugs

technical notes

   Some programs have comment starts inside quotes.  nocomequal
   is now smart enough to avoid changing these.

   In an earlier version of the progrm (before 2012 Nov 14) the
   comments were left as 'nibbons' that have no contents.  This allowed
   one to see exactly where the comments were.  One can remove these
   using Unix sed:

   set asterisk = '*'
   nocomequal < commented.p |\
   sed -e 's/{}//g' | \
   sed -e "s/(\$asterisk\$asterisk)//g" | \
   cat > commentless.p

   Since most of the time one doesn't want the extra empty, the
   program was changed to remove the nibbons.

*)
(* end module describe.nocomequal *)

(* trigger definitions follow to make program compilable.
The prgmod.p program has the triggers *)
(* module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* module filler.const from prgmod.p 4.20 *)

(* begin module interact.const *)
(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.48; (@ of prgmod.p 2011 May 03 *)
(* end module interact.const version = 5.48; (@ of prgmod.p 2011 May 03 *)

type

(* begin module interact.type *)
(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.48; (@ of prgmod.p 2011 May 03 *)
(* end module interact.type version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* module filler.type from prgmod.p 4.20 *)

(* module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* module trigger.type from prgmod.p 4.20 *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.clearstring *)
(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.48; (@ of prgmod.p 2011 May 03 *)
(* end module interact.clearstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module interact.writestring *)
(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.48; (@ of prgmod.p 2011 May 03 *)
(* end module interact.writestring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 5.48; (@ of prgmod.p 2011 May 03 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true.
1996 Sep 12: Bug found!  In the case of a trigger "ab", the program
used to miss it for situations like "aab".  This was because at the
first a it would step up.  Then it would see the second a and recognize
that was not part of ab.  It would fail to realize that it could be
the start of a new one.  The code now accounts for that possibility. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
{
writestring(list,seek);
writeln(list,'testfortrigger seek.letters[',state:1,']:',
seek.letters[state],' ch:',ch);
}
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin
            (* it failed.  But wait!  It could be the beginning
            of a NEW trigger string! *)
            if seek.letters[1] = ch then begin
               state := 1;
               skip := false;
               found := false
            end
            else begin (* reset trigger *)
               state := 0;
               skip := true;
               found := false
            end
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 5.48; (@ of prgmod.p 2011 May 03 *)

procedure themain(var fin, fout: text);
var
   c: char; (* the current character just read *)
   inquote: boolean; (* true if within a quote *)
   incomment1: boolean; (* true if within a parenthesis star comment *)
   incomment2: boolean; (* true if within a brackets comment *)
   p: char; (* the character previous to c *)
   internal: text; (* internal file *)
   comment1: trigger; (* trigger for comment type 1 *)
   comment2: trigger; (* trigger for comment type 2 *)
   inline: string; (* a line in the internal file *)
   gotten: boolean; (* inline was gotten *)
   outline: string; (* modified inline without nibbons *)
   done: boolean; (* done removing blanks *)

   equallines: boolean; (* the number of lines for input and output
      are made equal.  This preserves the number of lines to make
      p2c/tc translation easier to interpret. *)
begin
   equallines := true;

   (* Pass 1: remove the contents of the file *)
   incomment1 := false;
   incomment2 := false;
   c := ' ';
   rewrite(internal);
   while not eof(fin) do begin
      inquote := false; (* quotes can only be on one line *)
      while not eoln(fin) do begin
         p := c;
         read(fin,c);

         if not (incomment1 or incomment2)
         then if c = '''' (* This is a single quote mark in Pascal. *)
         then inquote := not inquote;

         if not (incomment1 or incomment2)
         then write(internal,c);

         if not inquote then begin (* skip comments inside quotes! *)
            if not incomment2 then begin
               if (p = '(') and (c = '*')
               then incomment1 := true;

               if (p = '*') and (c = ')') and incomment1
               then begin
                  incomment1 := false;
                  write(internal,'*)'); (* close of the comment *)
               end;
            end;
            if not incomment1 then begin
               if (c = '{')
               then incomment2 := true;
               if (c = '}') and incomment2
               then begin
                  incomment2 := false;
                  write(internal,'}'); (* close of the comment *)
               end;
            end
         end;
      end;

      readln(fin);
      if not (incomment1 or incomment2)
      then writeln(internal)
      else begin
        if equallines then writeln(internal);
      end

   end;

   (* Pass 2: remove the nibbons *)
   reset(internal);
   incomment1 := false;
   incomment2 := false;
   filltrigger(comment1,'(**)'); (* this is a nibbon *)
   filltrigger(comment2,'{}');   (* this is a nibbon *)
   while not eof (internal) do begin
      if eoln(internal) then begin
         readln(internal);
         (* The following if avoids adding an extra line at EOF! *)
         if not eof(internal) then writeln(fout);
      end
      else begin
         inquote := false; (* quotes can only be on one line *)
         (* the nibbons are only on one line *)
         resettrigger(comment1);
         resettrigger(comment2);
         clearstring(inline);
         clearstring(outline);
         getstring(internal, inline, gotten);

         (* step through inline, copy to outline *)
         while inline.current <= inline.length do begin
            c := inline.letters[inline.current];
            outline.length := outline.length + 1;
            outline.letters[outline.length] := c;

            if c = '''' (* This is a single quote mark in Pascal. *)
            then inquote := not inquote;

            if not inquote then begin (* skip comments inside quotes! *)
               testfortrigger(c, comment1);
               testfortrigger(c, comment2);

               if comment1.found then begin
                  (* back up the outline string!! *)
                  outline.length := outline.length - comment1.seek.length;
               end;

               if comment2.found then begin
                  (* back up the outline string!! *)
                  outline.length := outline.length - comment2.seek.length;
               end;
            end;
            inline.current := inline.current + 1;
         end;

         (* Remove extra blank spaces at the end of the line *)
         (* Note: this is done to all lines. *)
         done := false;
         while not done do begin
            if (outline.length = 0)
            then done := true
            else begin
               if outline.letters[outline.length] = ' ' then begin
                  outline.length := outline.length - 1;
               end
               else begin
                  done := true
               end
            end;
         end;

         if outline.length > 0 then begin
            writestring(fout, outline);
            writeln(fout);
         end
         else begin
            if equallines then writeln(fout);
         end;
      end
   end;
end;

begin
  themain(input, output);
1: end.
