program palinf(book, fout, palinfeatures, palinfp, output);
(* find palindromes, based on information theory
module libraries: delman, delmods
*)

label 1; (* end of program *)

const
(* begin module version *)
   version = 2.37; (* of palinf.p 2013 Jul 25
2013 Jul 25, 2.37: book piece names with '/' are broken.
2000 Jul 12, 2.36: upgraded names for both even and odd palindromes in
2000 Jul 12, 2.36: upgraded names for both even and odd palindromes in
                    palinfeatures, fixing an inconsistency from the previous
                    two name upgrades and adjusted the documentation
2000 Jul 11, 2.35: names upgraded to include piece name to make them
                    distinct.
1998 December 1: names upgraded to include bits to make them distinct.
1996 July 28: upgrade to produce palinfeatures
original: 1983 march 8 *)
(* end module version *)

(* begin module describe.palinf *)
(*
name
      palinf: find palindromes, based on information theory

synopsis
      palinf(book: in, palinfp: in,
             fout: out, palinfeatures: out, output: out)

files

      book: a book from the Delila system

      palinfp: parameters to control palinf, one per line

         1. The minimum rsequence of the palindrome to detect.
            alternatively, if the number is negative, it is the
            desired significance of the detected peaks, given in
            standard deviations.

         2. (Optional) size (integer).  The largest size palindrome allowed;
            base pairs across both halves of the site.  if omitted, the
            entire sequence is used (which may be very expensive).
            if this number is even, the next higher odd number will be used.

         3. (Optional) If the first character of this line is an 'm' then
            palinf will plot palindrome size (m) versus information content
            (rsequence).  A sharply rising curve indicates a good palindrome.
            'x' means plot position (x) versus information content (rsequence).
            a different character, such as 'n', means to list
            the detected palindromes.

      fout: Locations of palindromes.

         In the m mode, the coordinate location of significant palindromes
         (ie ones that passed the criterian) is given followed by a graph
         that shows the structure and significance of the palindrome from
         center to the outside:

 at position   725
                                   1                   2                   3
   m even  odd<0.1.2.3.4.5.6.7.8.9.0.1.2.3.4.5.6.7.8.9.0.1.2.3.4.5.6.7.8.9.0
   1 -0.5 -0.5=. 1 2 3 4 5         .         .         .         .         .
   2  1.0  1.0 . = 2 3 4 5         .         .         .         .         .
   3  2.5  0.5 .o 1 e2  3. 4  5    .         .         .         .         .
   4  4.0  0.0 o  1  2 e3. 4  5    .         .         .         .         .
   5  5.5 -0.5o.   1   2 .e3   4   5         .         .         .         .
   6  7.0 -1.0o.   1   2 . 3 e 4   5         .         .         .         .
   7  8.5  0.5 .o   1    2    3 e  4    5    .         .         .         .
   8  8.0  2.0 .   o1    2    3e   4    5    .         .         .         .
   9  7.5  3.5 .    1 o  2    e    4    5    .         .         .         .
  10  7.0  3.0 .    1o   2   e3    4    5    .         .         .         .
  11  6.5  4.5 .     1  o. 2e    3 .   4     5         .         .         .
  12  6.0  6.0 .     1   . =     3 .   4     5         .         .         .
  13  7.5  7.5 .     1   . 2  =  3 .   4     5         .         .         .
  14  7.0  9.0 .     1   . 2 e   o .   4     5         .         .         .
  15  8.5 10.5 .      1  .   2  e  .o      4 .    5    .         .         .
  16  8.0 12.0 .      1  .   2 e   .3  o   4 .    5    .         .         .
  17  9.5 13.5 .      1  .   2    e.3     o4 .    5    .         .         .
  18  9.0 15.0 .      1  .   2   e .3      4 o    5    .         .         .
  19  8.5 16.5 .       1 .     2e  .   3     . 4o      5         .         .
  20  8.0 18.0 .       1 .     e   .   3     . 4   o   5         .         .
  21  7.5 19.5 .       1 .    e2   .   3     . 4      o5         .         .
  22  7.0 21.0 .       1 .   e 2   .   3     . 4       5 o       .         .
  23  6.5 22.5 .       1 .  e  2   .   3     . 4       5    o    .         .
  24  8.0 22.0 .       1 .     e   .   3     . 4       5   o     .         .
  25  7.5 21.5 .        1.    e  2 .      3  .     4   .  o 5    .         .
  26  7.0 21.0 .        1.   e   2 .      3  .     4   . o  5    .         .
  27  8.5 20.5 .        1.      e2 .      3  .     4   .o   5    .         .
  28  8.0 20.0 .        1.     e 2 .      3  .     4   o    5    .         .
  29  7.5 19.5 .        1.    e  2 .      3  .     4  o.    5    .         .
  30  7.0 21.0 .        1.   e   2 .      3  .     4   . o  5    .         .
  31  6.5 20.5 .         1  e      2         3         4o        5         .
  32  6.0 22.0 .         1 e       2         3         4   o     5         .
  33  7.5 23.5 .         1    e    2         3         4      o  5         .
  34  7.0 25.0 .         1   e     2         3         4         o         .
  35  6.5 24.5 .         1  e      2         3         4        o5         .
 at   725           25.0 bits

      The horizontal axis is in bits, the vertical axis is in bases.  The
      numbers are the standard deviations.  With this chart one can determine
      the significance of each palindrome.  Clearly there is a strong (nearly
      standard deviations) odd palindrome at coordinate 725.

      In the x mode, the sequence is given:

                                     1                   2
    x bp even  odd<0.1.2.3.4.5.6.7.8.9.0.1.2.3.4.5.6.7.8.9.0.1.2.3.4.5.6.7.8
     2 a -0.5  1.5e. 1o2 3 4 5         .         .         .         .      
     3 c -0.5 -0.5=.  1  2  3. 4  5    .         .         .         .      
     4 a -0.5  3.0e.  1  o  3. 4  5    .         .         .         .      
     5 g -0.5  1.5e.  o1   2 . 3   4   5         .         .         .      
     6 t -0.5  1.5e.  o1   2 . 3   4   5         .         .         .      
     7 a  1.5  1.5 .  = 1    2    3    4    5    .         .         .      
     8 a -0.5  3.5e.    1 o  2    3    4    5    .         .         .      
     9 g  0.5 -0.5o.e   1    2    3    4    5    .         .         .      
    10 a -0.5  1.5e.  o 1    2    3    4    5    .         .         .      
    11 c -0.5  1.5e.  o  1   . 2     3 .   4     5         .         .      
    12 g  3.0  1.5 .  o  e   . 2     3 .   4     5         .         .      
    13 g  6.5 -0.5o.     1   . 2e    3 .   4     5         .         .      

      Here the horizontal axis is again in bits, but the vertical
      axis is the location on the sequence (which is why the bp column
      shows the bases).

      In the n mode, only a summary of the palindrome locations is provided:

           even      odd  palindromes
 at   537           21.0 bits
 at   547 24.5 bits
 at   707           22.5 bits
 at   725           25.0 bits
 at  1101 21.0 bits
 at  1180 24.0 bits
 at  1279           21.0 bits
 at  1322 24.5 bits

      palinfeatures: The locations of palindromes in the features format that
      the lister program uses.  Pass these to lister and the palindrome will
      be drawn on your sequence listing.

      The format that the features are listed is:

define "odd60.K00042" "-" "(((|)))" "(((|)))" -3 -2 -1 0 1 2 3
@ K00042 60 +1 "odd60.K00042" " 4.5 bits"

define "even547.K00042" "-" "((()))" "((()))" -3 -2 -1 0 1 2
@ K00042 547 +1 "even547.K00042" " 4.5 bits"

      output: messages to the user.

description

      Each piece of the book is searched for imperfect palindromes with
      significance determined by the first parameter in palinfp.  There are
      two kinds of palindrome: even and odd, refering to the size of the
      palindrome in bases.  An odd palindrome will have a central base, while
      an even one will not have one.  Method of use:  search without the 'm'
      option to pick out sites of interest.  Then use 'm' under 'stringent
      conditions' or on a smaller fragement to see the structure of the
      palindrome.  The final r value will be the maximum of r values for all
      smaller palindromes.  Note: equiprobable compositions are assumed for
      e(hnb).

      Theory:
      When there are a large number of sequences, the information needed to
      chose one of the 4 bases is log2(4) = 2 bits.  In contrast, for only
      two sequences (n = 2), the information measure is severely biased.
      This reflects the statistical likelyhood of finding matches.  One
      quarter of the time two randomly chosen bases will match.  In
      information theory terms, this means that a match counts only as 0.75
      bits (see reference Schneider1986 appendix figure A2).  So, for
      example, the restriction site for EcoRI, GAATTC is 6 x 2 = 12 bits when
      taken from many examples of the site (as when EcoRI binds).  However,
      as a single sequence, it only counts as 6 x 0.75 = 4.5 bits.  This
      effect prevents one from identifying spurious palindromes, but it is,
      unfortunately, not intuitive.

      To avoid duplicate definitions as much as possible, the names now
      include the piece name in which the palindrome is found.

examples

      The parameters

21 positive: bits minimum to find; negative: st.dev out to find
71 largest size palindrome to find (measured from center to edge in bases)
m  n=indicate detected palindromes; x=show sequence; m=show palindromes
   palinfp: parameters for palinf

      will locate the E. coli lac operator uniquely in the 401 bases
      surrounding the start of the lacZ transcript.

      The inverted repeats of pSC101 in GenBank K00042 are located with the
      same [13/35/m] parameters at coordinates 707 and 725.  (Other things
      are found as well, they have been ignored in the literature because
      they don't match the inverted repeats.)

      The parameters [4.5/6/n] will locate 6 base palindromes.

documentation

      Schneider, T.D., G.D. Stormo, L. Gold and A. Ehrenfeucht (1986)
      The information content of binding sites on nucleotide sequences.
      J. Mol. Biol. 188: 415-431.

see also

   {Example parameter file:} palinfp

   {Program to display the palindrome features:} lister.p

author

      Thomas D. Schneider and Karen A. Lewis

bugs

      If parameter 2 is very large, spurious sites will be found.

technical notes

      Limiting the size of the palindrome will increase the search speed.

*)
(* end module describe.palinf *)

(* begin module palinf.const *)
   fearraymax = 100; (* feature array max, largest number of palindrome
      matches that can be handled. *)
(* end module palinf.const *)

(* begin module book.const *)
(* constants needed for book manipulations *)
      dnamax = 1024;  (* length of dna arrays *)
      namelength = 100; (* maximum key name length *)
      linelength = 200; (* maximum line readable in book *)
(* end module book.const version = 7.75; {of delmod.p 2013 Jun 06} *)

type
(* begin module book.type *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);

      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;

(* begin module base.type *)
   (* define the four nucleotide bases *)
   base = (a,c,g,t);
(* end module base.type version = 7.75; {of delmod.p 2013 Jun 06} *)

      (* sequence types *)
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 7.75; {of delmod.p 2013 Jun 06} *)

var
      palinfp, (* parameters *)
      book, (* Delila system book *)
      fout, (* results *)
      palinfeatures: (* results *)
         text;

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 7.75; {of delmod.p 2013 Jun 06} *)

(* ************************************************************************ *)
(* end module package.primitive version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module package.getpiece *)
(* ************************************************************************ *)
(* begin module package.brpiece *)
(* ************************************************************************ *)
(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure writeline(var afile: text; l: lineptr;
                        carriagereturn: boolean);
(* write a line to a file, with carriage return if
carriagereturn is true. *)
var
   index: integer; (* index to characters in l *)
begin
   with l^ do begin
      for index := 1 to length
      do write(afile, letters[index]);
   end;
   if carriagereturn then writeln(afile);
end;

procedure showfreedna;
(* show the freedna list *)
var
   counter: integer; (* count of freedna list *)
   l: dnaptr; (* pointer into freedna list *)
begin
   l := freedna;
   counter := 0;
   while l <> nil do begin
      counter := succ(counter);
      write(output,counter:1);
      write(output, ', length = ',l^.length:1);
{
This is illegal according to gpc because one cannot
write a pointer to a text file.  It can be unearthed
for debugging.
      write(output, ', pointer id: ',l:1);
}
      writeln(output);
      l := l^.next
   end;
end;

procedure cleardna(var l: dnaptr);
(* clear the dna strutures to the free list *)
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function chomplement(b: char): char;
(* create the character complement of base b.  I must be getting hungry! *)
begin
  chomplement := basetochar(complement(chartobase(b)));
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
(* Note: the dirhomologous and dircomplement are treated as
plus and minus directions, which MIGHT NOT BE RIGHT! *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            dirhomologous,
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            dircomplement,
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
(* Note: the dirhomologous and dircomplement are treated as
plus and minus directions, which MIGHT NOT BE RIGHT! *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            dirhomologous,
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            dircomplement,
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.getto *)
function getto(var thefile: text; var theline: integer; ch: chset): char;
(* search the file for a character in the first line which is a
   member of the set ch.
Note:  on 1999 March 10 the definition of this function was cleaned
up.  Instead of putting thefile on the line AFTER the charcter ch
has been found, it puts thefile ON the line.  Other routines like
brdna and brpiece have to move to the next line themselves.  This makes
getto give the OBJECT.  *)
var
   achar: char; (* a character in thefile *)
   done: boolean; (* done finding achar *)
begin
   achar:=' ';
   done := false;
   while not done do begin
      if eof(thefile)
      then done := true
      else begin
         achar := thefile^;
         if achar in ch
         then done := true
         else begin
            readln(thefile);
            theline := succ(theline)
         end
      end
   end;
   if (achar in ch) then getto:=achar
                    else getto:=' '
{ The old method -
      while (not(achar in ch)) and (not eof(thefile))
         do begin
            readln(thefile,achar);
            theline := succ(theline)
         end;
   if (achar in ch) then getto:=achar
                    else getto:=' '
}
end;
(* end module book.getto version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.skipstar *)
procedure skipstar(var thefile: text);
(* skip start of line (or star = '*'). *)
begin (* skipstar *)
   if eof(thefile) then begin
      writeln(output,' procedure skipstar: end of book found');
      halt
   end
   else begin
      if thefile^ <> '*' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "*" expected as first character on the line, but "',
                        thefile^,'" was found');
         halt
      end;
      get(thefile); (* skip the star *)
      if thefile^ <> ' ' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "* " expected on a line but "*',
                        thefile^,'" was found');
         halt
      end;
      get(thefile) (* skip the blank *)
   end
end; (* skipstar *)
(* end module book.skipstar version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brreanum *)
procedure brreanum(var thefile: text; var theline: integer; var reanum: real);
(* read a real number from the file *)
begin
      skipstar(thefile);
      readln(thefile,reanum);
      theline := succ(theline)
end;
(* end module book.brreanum version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brnumber *)
procedure brnumber(var thefile: text; var theline: integer; var num: integer);
(* read a number from the file *)
begin
      skipstar(thefile);
      readln(thefile,num);
      theline := succ(theline)
end;
(* end module book.brnumber version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brname *)
procedure brname(var thefile: text; var theline: integer; var nam: name);
(* read a name from the file *)
var   i: integer; (* an index to the name *)
      c: char; (* a character read *)
begin (* brname *)
      skipstar(thefile);
      with nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(thefile,c);
            letters[length] := c
         until (eoln(thefile)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         if length<namelength
            then for i:=length+1 to namelength do letters[i]:=' '
      end;
      readln(thefile);
      theline := succ(theline)
end; (* brname *)
(* end module book.brname version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brline *)
procedure brline(var thefile: text; var theline: integer; var l: lineptr);
(* read a line from the file *)
var
      i: integer;
      acharacter: char;
begin
      skipstar(thefile);
      i:=0;
      (* protection added 2005 Sep 15 *)
      while (not eoln(thefile)) and (i < linelength)  do begin
         i:=succ(i);
         read(thefile,acharacter);
         l^.letters[i]:=acharacter
      end;
      (* protection added 2005 Sep 15 *)
      if not eoln(thefile) then begin
         writeln(output,'***********************************************');
         writeln(output,'* WARNING: brline: book line length exceeded');
         writeln(output,'* linelength > ',linelength:1,' characters');
         writeln(output,'* Only ',linelength:1,' characters read from book');
         writeln(output,'***********************************************');
      end;
      l^.length:=i;
      l^.next:=nil;
      readln(thefile);
      theline := succ(theline)
end;
(* end module book.brline version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brdirect *)
procedure brdirect(var thefile: text; var theline: integer; var direct: direction);
(* read a direction *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      theline := succ(theline);
      if ch='+' then direct:=plus
                else direct:=minus
end;
(* end module book.brdirect version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brconfig *)
procedure brconfig(var thefile: text; var theline: integer; var config: configuration);
(* read a configuration *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      theline := succ(theline);
      if ch='l' then config:=linear
                else config:=circular
end;
(* end module book.brconfig version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brnotenumber *)
procedure brnotenumber(var thefile: text; var theline: integer; var note: lineptr);
(* book note reading to obtain the number of the object.
the procedure returns the value of the number as a global.
(this is not such a good practice, but we are stuck with it for now.) *)
begin (* brnotenumber *)
      note:=nil;
      numbered := false;
      number := 0; (* force number to zero if there
         is no number at all *)
      (* the next character is n or * depending on whether there are notes *)
      if thefile^ = 'n' then begin
         readln(thefile);
         theline := succ(theline);
         if thefile^ <> 'n' then begin
            skipstar(thefile);
            if not eoln(thefile) then begin
               if thefile^ = '#' then begin
                  numbered := true;
                  get(thefile); (* move past the number symbol *)
                  read(thefile,number);
               end
            end;
            repeat
               readln(thefile);
               theline := succ(theline)
            until thefile^ = 'n';
            readln(thefile);
            theline := succ(theline)
         end
         else begin
            readln(thefile);
            theline := succ(theline)
         end
      end
end; (* brnotenumber *)
(* end module book.brnotenumber version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brnote *)
procedure brnote(var thefile: text; var theline: integer; var note: lineptr);
(* read note key *)
var
      newnote: lineptr; (* the new note *)
      previousnote: lineptr; (* the last line of the notes *)
begin (* brnote *)
      note:=nil;
      if thefile^ = 'n' then begin (* enter note *)
         readln(thefile);
         theline := succ(theline);
         if thefile^ <> 'n' then begin (* abort null note (n/n) *)
            getline(note);
            newnote:=note;
            while thefile^ <> 'n' do begin (* wait until end of note *)
               brline(thefile,theline,newnote);
               previousnote:=newnote;
               (* get next note *)
               getline(newnote^.next);
               newnote:=newnote^.next;
            end;
            (* last note was not used, so: *)
            clearline(newnote);
            previousnote^.next:=nil;
            readln(thefile);
            theline := succ(theline);
         end
         else begin
            readln(thefile);
            theline := succ(theline);
         end;
      end
end; (* brnote *)
(* end module book.brnote version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brheader *)
procedure brheader(var thefile: text; var theline: integer; var hea: header);
(* read the header of a key. *)
begin
      with hea do begin
         readln(thefile); (* move past the object name - new definition 1999 Mar 13 *)
         theline := succ(theline);
{bbb}
         (* read key name *)
         brname(thefile,theline,keynam);

         (* read full name *)
         getline(fulnam);
         brline(thefile,theline,fulnam);

         (* read note key *)
         if readnumber then brnotenumber(thefile,theline,note)
                       else brnote(thefile,theline,note)
      end
end;
(* end module book.brheader version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.copyheader *)
procedure copyheader(fromhea: header; var tohea: header);
(* copy the header fromhea into tohea.  Note that the
linked objects are NOT copied, but merely pointed to. *)
begin
   tohea.keynam.letters := fromhea.keynam.letters;
   tohea.keynam.length  := fromhea.keynam.length;
   tohea.note           := fromhea.note;
   tohea.fulnam         := fromhea.fulnam;
end;
(* end module book.copyheader version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brpiekey *)
procedure brpiekey(var thefile: text; var theline: integer; var pie: piekey);
(* read piece key, track the line number *)
begin
      with pie do begin
         brheader(thefile,theline,hea);
         brreanum(thefile,theline,mapbeg);
         brconfig(thefile,theline,coocon);
         brdirect(thefile,theline,coodir);
         brnumber(thefile,theline,coobeg);
         brnumber(thefile,theline,cooend);
         brconfig(thefile,theline,piecon);
         brdirect(thefile,theline,piedir);
         brnumber(thefile,theline,piebeg);
         brnumber(thefile,theline,pieend);
      end
end;
(* end module book.brpiekey version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brdna *)
procedure brdna(var thefile: text; var theline: integer; var dna: dnaptr);
(* read in dna from thefile, track the line *)
(* note: if the dna were circularized, by linking the last dnastring
to the first, then the cleardna routine could not clear properly,
and would loop forever... there is no reason to do that, since a simple
mod function will allow one to access the circle. *)
var
      ch: char;
      workdna: dnaptr;
begin
      getdna(dna);
      workdna:=dna;
      ch:=getto(thefile,theline,['d']);
      readln(thefile);
      theline := succ(theline);
      read(thefile,ch); (* skipstar *)
      while (ch = '*') do
      begin
         read(thefile,ch); (* skip blank *)
         repeat
            read(thefile,ch);
            if ch in ['a','c','g','t'] then begin
               if workdna^.length=dnamax then begin
                  getdna(workdna^.next);
                  workdna:=workdna^.next
               end;
               workdna^.length:=succ(workdna^.length);
               workdna^.part[workdna^.length]:=chartobase(ch)
            end
         until eoln(thefile);
         readln(thefile); (* go to next line *)
         theline := succ(theline);
         read(thefile,ch); (* ch is either '*' or 'd' *)
      end;
      readln(thefile); (* read past the d *)
      theline := succ(theline);
end;
(* end module book.brdna version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brpiece *)
procedure brpiece(var thefile: text; var theline: integer; var pie: pieceptr);
(* read in a piece, change theline to reflect the lines traversed *)
begin
{
   readln(thefile); (* move past the word 'piece' - new definition 1999 Mar 13 *)
   theline := succ(theline); (* BUG: was below! *)
bbb}
   brpiekey(thefile,theline,pie^.key);
   if numbered or (not skipunnum)
   then brdna(thefile,theline,pie^.dna);
   readln(thefile); (* move past the word 'piece' - new definition 1999 Mar 13 *)
   theline := succ(theline);
end;
(* end module book.brpiece version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.brinit *)
procedure brinit(var book: text; var theline: integer);
(* check that the book is ok to read, and
   set up the global variables for br routines *)
begin (* brinit *)
      (* halt if the book is bad (first word is 'halt') or the first
         character is not * *)
      reset(book);
      if not eof(book) then begin
         (* check for the date line *)
         if book^ <> '*' then begin
            if book^ <> 'h'
            then writeln(output, ' this is not the first line of a book:')
            else writeln(output, ' bad book:');
            write(output, ' ');

            while not (eoln(book) or eof(book)) do begin
               write(output, book^);
               get(book)
            end;
            writeln(output);
            halt
         end
      end
      else begin
         writeln(output, ' book is empty');
         halt
      end;

      (* initialize free storage *)
      freeline:=nil;
      freedna:=nil;

      readnumber:=true; (* usually we read in numbers for items *)
      number:=0; (* arbitrary value *)
      numbered:=false; (* the piece has no number (none yet read in) *)
      skipunnum:=false;
   theline := 1;
end; (* brinit *)
(* end module book.brinit version = 7.75; {of delmod.p 2013 Jun 06} *)

(* ************************************************************************ *)
(* end module package.brpiece version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.getpiece *)
procedure getpiece(var thefile: text; var theline: integer; var pie: pieceptr);
(* move to and read in the next piece in the book *)
var   ch: char;
begin
   ch:=getto(thefile,theline,['p']); (* get to the next p(iece) in the book *)
   if ch<>' ' then begin
      brpiece(thefile,theline,pie);
{
1999 june 2: removed this:
      ch:=getto(thefile,theline,['p']); (* read to end of p *)
}
{
bbb - now done in brpiece
      readln(thefile); (* read past piece *)
      theline := succ(theline);
}
   end
   else clearpiece(pie);
end;
(* end module book.getpiece version = 7.75; {of delmod.p 2013 Jun 06} *)

(* ************************************************************************ *)
(* end module package.getpiece version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module book.getbase *)
function getbase(position: integer; pie: pieceptr):base;
(* Get a base from the position (internal coordinates) of the
   piece.  Protection is made against positions outside the piece.
In the case of circles it would be convenient to wrap around when
requests are off the end.  So the routine will do a modular
wrap for positions outside the range 1 to the length.
This is a new feature as of 2000 March 22.
*)
var
   workdna: dnaptr; (* pointer to the dna part of pie *)
   p: integer; (* current count of bases into the workdna *)
   spot: integer; (* the last base of the dna part *)
   thelength: integer; (* the length of the piece *)
begin
{
writeln(output,'NEW getbase: position=',position:1,'^^^^^^^^^^^^^^^^^^^^');
}
   (* handle cases of position out of range by circular wrapping *)
   thelength := piecelength(pie);
   while position < 1         do position := position + thelength;
   while position > thelength do position := position - thelength;

   workdna:=pie^.dna;
   p:=workdna^.length;
   while position > p do begin
{
writeln(output,' workdna^.length=',workdna^.length:1);
}
      workdna := workdna^.next;
      if workdna = nil then begin
        writeln(output,'error in function getbase!');
        halt
      end;
      p := p + workdna^.length;
   end;
{
writeln(output,'p=',p:1);
}
   if workdna = nil then begin
      writeln(output,'error in getbase: request off end of piece');
      halt
   end
   else begin
      spot := workdna^.length - (p-position);
{
writeln(output,'spot=',spot:1);
showdnasegment(output,workdna, spot);
}
      if (spot <= 0) then begin
         writeln(output,'error in getbase, spot (= ',spot:1, ') must be positive');
         halt
      end;
      if (spot > workdna^.length) then begin
         writeln(output,'error in getbase, spot (=',spot:1,
                        ') must be less than length (=',workdna^.length:1,')');
         halt
      end;
{
writeln(output,'base = ', workdna^.part[spot]);
}
      getbase:=workdna^.part[spot]
   end
end;
(* end module book.getbase version = 7.75; {of delmod.p 2013 Jun 06} *)

(* begin module missparam *)
procedure missparam(var param: text);
(* look at param to see if the next parameter is missing
this is useful when reading in a series of parameters.  use it
just before readln of each parameter.*)
begin (* missparam *)
      if eof(param) then begin
         writeln(output,' missing parameter');
         halt
      end
end; (* missparam *)
(* end module missparam version = 7.75; {of delmod.p 2013 Jun 06} *)

procedure initialize(var palinfp, book, fout, palinfeatures: text;
                     var sig: real; var boundary: integer;
                     var mmode, xmode: boolean;
                     var apiece: pieceptr; theline: integer);
(* start things up (see procedure themain for definitions *)
var
      maxsize: integer; (* largest palindrome allowed *)
begin (* initialize *)
      writeln(output,'palinf ',version:4:2);
      theline := 0;

      rewrite(fout);
      writeln(fout,'palinf ',version:4:2,' palindrome search',
                   ' based on information theory');
      brinit(book, theline);
      write(fout,' ');
      copyaline(book,fout);
      writeln(fout);

      rewrite(palinfeatures);
      writeln(palinfeatures,'* palinf ',version:4:2,' palindrome search',
                   ' based on information theory');
      brinit(book, theline);
      copyaline(book,palinfeatures);

      (* get parameters *)
      reset(palinfp);
      missparam(palinfp);
      readln(palinfp, sig);

      if eof(palinfp)
      then boundary := maxint
      else begin
         readln(palinfp, maxsize);

         if maxsize < 1 then begin
            writeln(output,'Maximum size must be positive');
            halt
         end;

         case maxsize mod 2 of (* is maxsize odd or even *)
            0: boundary := round(maxsize/2);     (* even *)
            1: boundary := round((maxsize-1)/2); (* odd *)
         end;

         mmode := false;
         xmode := false;
         if not eof(palinfp)
         then if palinfp^ = 'm' then mmode := true
         else if palinfp^ = 'x' then xmode := true;
      end;

      (* display parameters *)
      if sig >= 0.0
      then writeln(fout,' ',sig:8:2,' bits is the lowest',
             ' rsequence detected')
      else begin
         writeln(fout,' ',sig:8:5,' palindromes of this significance');
         writeln(fout,'    (in standard deviations) or more are found');
      end;
      writeln(fout,' ',(2*boundary + 1):8,' largest palindrome',
         ' found (bases)');

      if not (mmode or xmode)
      then writeln(fout,' not in a graphic mode.')
      else if mmode then writeln(fout,
         ' m versus rsequence is plotted for each detected palindrome')
      else if xmode then writeln(fout,
         ' x versus rsequence is plotted for all positions (x)');
      writeln(fout);

      if mmode then begin
         writeln(fout,' m is the distance from the center',
                      ' of the odd palindrome.');
         writeln(fout)
      end;

      if mmode or xmode then begin
         writeln(fout,' the numbers on the plot are',
               ' standard deviations of e(hnb) for');
         writeln(fout,' comparing the palindrome to expected values.');
         writeln(fout);
      end;

      writeln(fout,' warning:');
      writeln(fout,' the information content is found by summing from');
      writeln(fout,' the center to the edges of the palindrome.');
      writeln(fout,' since the highest sum found is used for detection,');
      writeln(fout,' palindromes may appear more significant than');
      writeln(fout,' they really are.');
      writeln(fout);

      writeln(fout,' an odd palindrome is centered on the base x;');
      writeln(fout,' while an even one is centered just before the base x,',
                   ' between bases.');

      new(apiece)
end; (* initialize *)

procedure showpiece(var fout: text; apiece: pieceptr);
(* show some piece information to fout *)
begin (* showpiece *)
      writeln(fout);
      if number > 0 then write(fout,' piece number: ',number:1);

      with apiece^.key.hea.keynam
         do write(fout,'  name: ',letters:length);

      with apiece^.key.hea.fulnam^
         do writeln(fout,'  full name: ',letters:length);
end; (* showpiece *)

procedure palsearch(var fout, palinfeatures: text;
                    pie: pieceptr;
                    sig: real;
                    boundary: integer; mmode,xmode: boolean);
(* Search a piece pie for palindromes.  See procedure themain for
other definitions *)
const
   infofield = 4; (* width of information field display *)
   infodecim = 1; (* decimal places in infofield *)
   defmaxr = 30; (* default for maxr: maximum bits plottable *)
var
   ehnb, (* e(hnb) for n=2 *)
   varhnb, (* the variance of e(hnb) when n = 2 *)
   equal, (* rsequence for match of two bases *)
   unequal, (* rsequence for mismatch of two bases *)
   halfunequal, (* half of unequal - used to initialize odd palindromes *)
   reven, (* rsequence for even palindromes *)
   rodd, (* rsequence for odd palindromes *)
   maxreven, (* the maximum value of reven at one l *)
   maxrodd,  (* the maximum value of rodd  at one l *)
   minrs: (* the minimum information content site to display *)
      real;
   graphics: boolean; (* whether graphs are being made or not *)
   maxr, (* maximum bits plottable *)
   n, (* size of the piece (internal coordinates) *)
   x, (* location in the piece (internal coordinates) *)
   m, (* number of bases on one flank of the palindrome.
      if the palindrome is even, its size is 2*m,
      if the palindrome is odd, its size is 2*m+1. *)
   xm: (* x + m *)
      integer;
   b1, b2, b3: base; (* bases at several positions in the sequence.
      b1 and b2 are a pair in an even palindrome;
      b1 and b3 are a pair in an odd palindrome. *)
   twomaxr: integer; (* twice maxr, for graphics *)

   evenmatch: array[1..fearraymax] of integer; (* feature array.  This
      contains the locations of even palindrome matches *)
   evennumber: integer; (* number of matches in evenmatch *)

   oddmatch: array[1..fearraymax] of integer; (* feature array.  This
      contains the locations of odd palindrome matches *)
   oddnumber: integer; (* number of matches in oddmatch *)

procedure graph10s(var fout: text; maxr: integer);
(* put digits every 10 positions up to maxr in the file fout for a graph *)
var
      r: integer; (* index for counting in the graph *)
begin (* graph10s *)
      for r := 1 to maxr div 10
         do write(fout,' ':19,r:1);
      writeln(fout);
end; (* graph10s *)
procedure graph1s(var fout: text; maxr: integer);
(* graph the tick marks for every 1 (and half) bit positions to file fout;
maxr is the maximum to plot *)
var
      r: integer; (* index for counting in the graph *)
begin (* graph1s *)
      write(fout,'<0');
      for r := 1 to maxr
         do write(fout,'.',(r mod 10):1);
      writeln(fout)
end; (* graph1s *)
procedure graphline(var fout: text; reven, rodd, stdev: real;
                     twomaxr: integer);
(* graph a line of a plot to file fout.  points for reven and rodd are
made.  maxr is the largest r accepted, and twomaxr is twice that.
stdev is the standard deviation in bits. *)
var
      r, (* index to position in the graph *)
      tworeven, (* reven made integer *)
      tworodd:  (* rodd  made integer *)
         integer;
      twostdev: real; (* twice the standard deviation, for plotting *)
begin (* graphline *)
      (* handle negative r values *)
      if (reven < 0) or (rodd < 0)
      then begin
         if (reven < 0) and (rodd < 0)
         then write(fout,'=')
         else if rodd < 0
              then write(fout,'o')
              else write(fout,'e')
      end
      else write(fout,' ');

      tworeven := round(2*reven);
      if tworeven > twomaxr then tworeven := twomaxr;
      tworodd  := round(2*(rodd - halfunequal));
      if tworodd  > twomaxr then tworodd  := twomaxr;

      twostdev := round(2*stdev);

      for r := 0 to twomaxr
      do if (tworeven = r) and (tworodd = r) then write(fout,'=')
      else if tworeven = r then write(fout,'e')
      else if tworodd  = r then write(fout,'o')
      else if     twostdev = r then write(fout,'1')
      else if 2 * twostdev = r then write(fout,'2')
      else if 3 * twostdev = r then write(fout,'3')
      else if 4 * twostdev = r then write(fout,'4')
      else if 5 * twostdev = r then write(fout,'5')
      else if (((round(r/2) mod 10) = 0) or
               ((round(r/2) mod  5) = 0)) and (not odd(r))
           then write(fout,'.')
           else write(fout,' ');
      writeln(fout)
end; (* graphline *)
procedure space(var f: text; s: integer);
(* send s spaces to the file f.  this gets around pascal sending spaces
when s is <= 0.  *)
var
      count: integer; (* counts the spaces *)
begin (* space *)
      if s > 0 then for count :=1 to s do write(f,' ');
end; (* space *)
procedure mrgraphheader;
(* create the header of a graph of m versus r *)
begin (* mrgraphheader *)
      (* position on sequence *)
      writeln(fout,' at position ',inttopie(x,pie):5);

      (* move over to edge of graph *)
      space(fout,2*infofield+8);

      (* put digits for every 10 *)
      graph10s(fout, maxr);

      (* define each column *)
      space(fout,3);
      write(fout,'m');
      space(fout,infofield - 3);
      write(fout,'even');
      space(fout,infofield - 2);
      write(fout,'odd');

      (* begin the edge of the graph *)
      graph1s(fout, maxr);
end; (* mrgraphheader *)
procedure mrgraphic;
(* show r values for all m *)
begin (* mrgraphic *)
(*      write(fout,' ',inttopie(l,pie):5); *)
      write(fout,' ',m:3);
      write(fout,' ',reven:infofield:infodecim);
      write(fout,' ',rodd:infofield:infodecim);
      graphline(fout,reven,rodd,2*sqrt(varhnb*m),twomaxr)
end; (* mrgraphic *)
procedure xrgraphheader(var fout: text; maxr: integer);
(* create the header of a graph of x versus r *)
begin (* xrgraphheader *)
      (* move over to edge of graph *)
      space(fout,2*infofield+10);

      (* put digits for every 10 *)
      graph10s(fout, maxr);

      (* define each column *)
      space(fout,4);  (* one space devoted to 'bp' below *)
      write(fout,'x');
      write(fout,' bp');
      space(fout,infofield - 3);
      write(fout,'even');
      space(fout,infofield - 2);
      write(fout,'odd');

      (* begin the edge of the graph *)
      graph1s(fout, maxr);
end; (* xrgraphheader *)
procedure xrgraphic(var fout: text; x: integer; b: char;
                    maxreven, maxrodd: real;
                    twomaxr: integer);
(* produce a graph line of x position versus r for both even and odd
palindromes.  maxr is the maximum graph, and twomaxr is twice that.
note that the global m will be the largest of the palindrome, and
so can be used for finding the standard deviation.
b is the base (as a character) corresponding to position l *)
begin (* xrgraphic *)
      write(fout,' ',inttopie(x,pie):5);
      write(fout,' ',b);
      write(fout,' ',maxreven:infofield:infodecim);
      write(fout,' ',maxrodd:infofield:infodecim);
      graphline(fout,maxreven,maxrodd,2*sqrt(varhnb*m),twomaxr);
end; (* xrgraphic *)

procedure maximize(r: real; var maxr: real);
(* maxr is set to the maximum of all r *)
begin (* maximize *)
      if r > maxr then maxr := r
end; (* maximize *)

procedure match(evenodd: char; var r: real; left, right: base);
(* calculate the information content for matching (or not) of
two bases on the left and right side of the palindrome.
sum into r.  evenodd is either 'e' (even) or 'o' (odd). *)
begin (* match *)
   if left = complement(right)
   then begin
      r := r + equal;

      (* record the location of the match in the proper array *)
      case evenodd of
        'e': begin
           evennumber := succ(evennumber);
           if evennumber > fearraymax then begin
              writeln(output,'An even palendrome has more than ',
                             fearraymax:1, ' matches.');
              writeln(output,'increase constant fearraymax');
              halt;
           end;
           evenmatch[evennumber] := m
        end;
        'o': begin
           oddnumber := succ(oddnumber);
           if oddnumber > fearraymax then begin
              writeln(output,'An odd palendrome has more than ',
                             fearraymax:1, ' matches.');
              writeln(output,'increase constant fearraymax');
              halt;
           end;
           oddmatch[oddnumber] := m
        end;
      end;
   end
   else r := r + unequal
end; (* match *)

function within(x,m,n: integer): boolean;
(* are the locations x-m and x+m within the range 1 to n *)
begin (* within *)
      within := (x-m >= 1) and (x+m <= n)
end; (* within *)

procedure dolinear(showit: boolean);
(* handle the linear case, and perhaps show the intermediate results *)
begin (* dolinear *)
      while within(x,m,n) and (m <= boundary) do begin
         xm := x + m;

         b1 := getbase(x-m, pie);
         b2 := getbase(xm-1, pie);
         b3 := getbase(xm, pie);

         match('e',reven, b1, b2);
         match('o',rodd,  b1, b3);
         maximize(reven,maxreven);
         maximize(rodd,maxrodd);
         if showit then mrgraphic;
         m := m + 1
      end
end; (* dolinear *)

function wrap(p, n: integer): integer;
(* if p is outside the range 1 to n,
wrap it back into the range so that 1 to n becomes circular *)
begin (* wrap *)
      if p < 1
      then wrap := n + p (* nb: this will not handle all p *)
      else if p > n
           then wrap := ((p-1) mod n) + 1
           else wrap := p
end; (* wrap *)

procedure docircular(showit: boolean);
(* handle the circular case, and perhaps show the intermediate results *)
begin (* docircular *)
      while m <= boundary do begin
         xm := x + m;

         b1 := getbase(wrap(x-m, n), pie);
         b2 := getbase(wrap(xm-1, n), pie);
         b3 := getbase(wrap(xm, n), pie);

         match('e',reven, b1, b2);
         match('o',rodd,  b1, b3);
         maximize(reven,maxreven);
         maximize(rodd,maxrodd);
         if showit then mrgraphic;
         m := m + 1
      end
end; (* docircular *)

procedure search(showit: boolean);
(* search for a palindrome at location x,
and perhaps show the intermediate results *)
begin (* search *)
      reven := 0.0;
      rodd  := halfunequal;   (* since m starts at 1 (below), the central
         position of an odd palindrome (which never is the complement of
         itself) must be preset *)
      maxreven := -maxint; (* maxima must start with the smallest value
                  possible *)
      maxrodd := -maxint;
      m := 1; (* begin with smallest palindrome possible *)
      evennumber := 0;
      oddnumber := 0;

      if showit then mrgraphheader;

      (* search *)
      case pie^.key.piecon of (* topology question *)
         linear: dolinear(showit);
         circular: docircular(showit)
      end
end; (* search *)


procedure recordpalinfeatures(var pf: text; apiece: pieceptr);
(* record the current even or odd palindrome into the pf file *)
var
   mindex: integer; (* index to palindrome matches *)
   rpt: integer; (* repeat an operation *)
begin

{ 
The upgrade of 1998 December 1 made the names written to palinfeatures unique
by adding the information content of the palindrome into the name and
removing the following display of information content.  The resulting
name of the palinfeature looked like this:
 
"odd60-7.3-bits"  
 
However, this was only done for *odd* palinfeatures, not even ones.  The even
palindromes still contained only the kind and size of the palindrome (e.g.
even547) in the name and displayed the information content as a separate
entity in the features.  The upgrade of 2000 July 12 changed the name of both
odd and even palindromes to be the kind, size, and piece name (e.g.
odd60.K00042) and then displays the information content as a separate
entity.  
}

   if maxreven >= minrs then begin (* significance test *)
{
      writeln(pf);
      writeln(pf,'* at ',inttopie(x, pie):1);
      write(pf, maxreven:infofield:infodecim,' bits');
      write(pf,'* even: ');
      for mindex := 1 to evennumber
      do write(pf,' ', evenmatch[mindex]:1);
      writeln(pf);
}


      (* define "odd60.K00042" "-" "(((|)))" "(((|)))" -3 -2 -1 0 1 2 3 *)
{
old method (prior to 1998 December 1) without having bits or piece name as
part of name:
      write(pf, 'define "even',inttopie(x, pie):1,'" "-"');
}
      write(pf, 'define "even',inttopie(x, pie):1,'.');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,'" "-"');
      for rpt := 1 to 2 do begin
         write(pf, ' "');
         for mindex := 1 to evennumber do write(pf,'(');
         for mindex := 1 to evennumber do write(pf,')');
         write(pf,'"');
      end;
      for mindex := evennumber downto 1 do write(pf,' ',-evenmatch[mindex]:1);
      (* there is no central base *)
      for mindex := 1 to evennumber do write(pf,' ',(+evenmatch[mindex]-1):1);
      writeln(pf);

      (* @ K00042 60 +1 "odd60.K00042" " 4.5 bits" *)
      write(pf,'@ ');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,' ',inttopie(x, pie):1);
{
old method (prior to 1998 December 1) without having bits or piece name as
part of name:
      write(pf,' +1 "odd',inttopie(x, pie):1,'"');
      write(pf,' "',maxrodd:infofield:infodecim,' bits"');
}
      write(pf,' +1 "even',inttopie(x, pie):1,'.');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,'" "',maxreven:infofield:infodecim,' bits"');
      writeln(pf);
   end;

   if maxrodd >= minrs then begin (* significance test *)
      writeln(pf);
{
      writeln(pf,'* at ',inttopie(x, pie):1);
      write(pf,' ',maxrodd:infofield:infodecim,' bits');
      write(pf,'* odd: ');
      for mindex := 1 to oddnumber
      do write(pf,' ', oddmatch[mindex]:1);
      writeln(pf);
}

      (* define "even547.K00042" "-" "((()))" "((()))" -3 -2 -1 0 1 2 *)
{
old method (prior to 1998 December 1) without having bits or piece name as
part of name:
      write(pf, 'define "odd',inttopie(x, pie):1,'" "-"');
}
      write(pf, 'define "odd',inttopie(x, pie):1,'.');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,'" "-"');
      for rpt := 1 to 2 do begin
         write(pf, ' "');
         for mindex := 1 to oddnumber do write(pf,'(');
         write(pf,'|');
         for mindex := 1 to oddnumber do write(pf,')');
         write(pf,'"');
      end;

      for mindex := oddnumber downto 1 do write(pf,' ',-oddmatch[mindex]:1);
      write(pf,' ',0:1); (* the central base *)
      for mindex := 1 to oddnumber do write(pf,' ',+oddmatch[mindex]:1);
      writeln(pf);

      (* @ K00042 547 +1 "even547.K00042" " 4.5 bits" *)
      write(pf,'@ ');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,' ',inttopie(x, pie):1);
{
old method (prior to 1998 December 1) without having bits or piece name as
part of name:
      write(pf,' +1 "odd',inttopie(x, pie):1,'"');
      write(pf,' "',maxrodd:infofield:infodecim,' bits"');
}
      write(pf,' +1 "odd',inttopie(x, pie):1,'.');
      with apiece^.key.hea.keynam
         do write(pf,letters:length);
      write(pf,'" "',maxrodd:infofield:infodecim,' bits"');
      writeln(pf);

   end;

end;

begin (* palsearch *)
      ehnb := 0.75; (* for n=2 and equiprobable composition *)
      varhnb := 0.1875; (* as with ehnb *)

      (* precalculate rsequencel = e(hnb) - hy(x).
         the factor of two accounts for the two halves of the
         palindrome.  the subtracted values can easily be shown to be
         the values for hy(x) for 2 bases, matching or not *)
      equal := 2*(ehnb - 0.0);
      unequal := 2*(ehnb - 1.0);
      halfunequal := 0.0; (* for the center of an odd palindrome *)
         (* this will be ignored in this version of palinf.  it would be
            ehnb - 1.0 if one wanted to count the central base as not
            complementing itself *)

      if sig >= 0 then begin
         minrs := sig;
         writeln(fout,' minimum rsequence = ',minrs:infofield:infodecim,
                      ' the minimum information content to display')
      end
      else minrs := maxint;

      writeln(fout);
      graphics := mmode or xmode;
      if not graphics
      then writeln(fout,' ':(infofield+7),'even',
                        ' ':(infofield+2), 'odd  palindromes');

      n := piecelength(pie);

      (* prepare header of x versus r graph *)
      if xmode then begin
         maxr := defmaxr - 2; (* xmode plot is wider than mmode since the
            the number x is in general bigger than m, so the maximum of the
            plot must be less *)
         xrgraphheader(fout, maxr)
      end
      else maxr := defmaxr;
      twomaxr := 2*maxr; (* to speed up graphics *)

      (* x starts at 2 and ends at n-1 since no palindrome centered at
         1 or n can be detected *)
      for x := 2 to n-1 do begin (* search the piece *)
         search(false);

         (* display *)
         if xmode
         then xrgraphic(fout,x,basetochar(getbase(x,pie)),
                        maxreven,maxrodd,twomaxr)
         else begin (* do mmode and no graphics *)
            if sig < 0 then minrs := -sig * 2 * sqrt(varhnb * m);
               (* m is the largest palindrome found in the first search
                  so it can be used to calculate the st. dev. *)

            if (maxreven >= minrs) or
               (maxrodd  >= minrs) then begin

               (* repeat the search, showing the values at each m: *)
               if mmode then search(true);

               write(fout,' at ',inttopie(x, pie):5,' ');

               if maxreven >= minrs
               then write(fout, maxreven:infofield:infodecim,' bits')
               else write(fout,' ':(infofield+5));

               if maxrodd >= minrs
               then write(fout,' ',maxrodd:infofield:infodecim,' bits');
               if mmode then writeln(fout); (* space out displays *)
               writeln(fout);

               recordpalinfeatures(palinfeatures, pie);

            end
         end
      end
end; (* palsearch *)

procedure themain(var palinfp, book, fout, palinfeatures: text);
(* main procedure: search book for palindromes, list to fout, based
on parameters in palinfp *)
var
   sig: (* if positive, significance of spikes in bits
           if negative, significance in standard deviations
           of palindromes to display *)
      real;
   boundary: integer; (* 2*boundary + 1 (in bp) is the largest
      palindrome found *)
   mmode, (* show an m versus rsequence plot at each detected palindrome *)
   xmode: (* show an l versus rsequence plot *)
      boolean;
   apiece: pieceptr; (* pointer to a sequence piece *)
   theline: integer; (* book line number *)
begin (* themain *)
      initialize(palinfp, book, fout, palinfeatures,
                 sig,boundary,mmode,xmode,apiece, theline);

      (* start reading pieces *)
      while not eof(book) do begin
         getpiece(book, theline, apiece);
         if not eof(book) then begin
            showpiece(fout, apiece);
            palsearch(fout, palinfeatures, apiece, sig, boundary, mmode, xmode);
            clearpiece(apiece)
         end
      end
end; (* themain *)

begin (* palinf *)
      themain(palinfp, book, fout, palinfeatures);
1: end. (* palif *)
