program parseform(input, {parseformp,} output);
(* parseform: parse an html form

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.05; (* of parseform.p 2016 Jan 29
2016 Jan 29: 1.05: debug for p2c/tc get(input); -> read(input,c);
2000 Oct 16: 1.04: allow %22 for quotes
2000 Jun 29: 1.03: fix hex conversion
origin 1999 Apr 21 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.parseform *)
(*
name
   parseform: parse an html form

synopsis
   parseform(input: in, parseformp: in, output: out)

files
   input:  data string from a <FORM> on an html page.

   parseformp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      NOT USED

   output: parsed form

description
   
   An input string from an html form is parsed into lines.
   & symbols represent line breaks.
   + symbols represent spaces
   % begin a hex representation of a character, %2b is '+'
   the variable name is placed after the string data

examples

For the input:

variable-name=initial-value&bread=rye&meat=MeatBall&comments=my+text&storeNum=230432  

the output is:

initial-value variable-name
rye bread
MeatBall meat
my text comments
230432 storeNum

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.parseform *)

{
var
   parseformp: (* file used by this program *)
   text; (* file used by this program *)
}

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module decapitalize *)
function decapitalize(c: char): char;
(* convert the character c to lower case *)
var
   n: integer; (* c is the n'th letter of the alphabet *)
begin
   n := ord(c);

   if (n >= ord('A')) and
      (n <= ord('Z'))
   then c := chr( n - ord('A') + ord('a'))
   else c := chr(n);

   decapitalize := c
end;
(* end module decapitalize version = 4.21; (@ of prgmod.p 1997 October 22 *)

(* begin module parseform.themain *)
procedure themain; {(var parseformp: text);}
(* the main procedure of the program *)
const
   showoriginal = 2; (* how to show the data *)
   maxstring = 200; (* maximum datatype string length *)
var
   c: char; (* a character read from input *)
   inhex: boolean; (* whether we are inside or outside a hexidecimal character *)
   inname: boolean; (* whether we are inside or outside a name *)
   hexnumber: integer; (* a hexidecimal number *)
   hexcount: integer; (* number of digits so far in a hexidecimal number *)
   n: integer; (* index to typestring *)
   namelength: integer; (* current length of name *)
{
   parameterversion: real; (* parameter version number *)
}
   typestring: array[1..maxstring] of char;
begin
{
   writeln(output,'parseform ',version:4:2);
   reset(parseformp);
   readln(parseformp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

   if showoriginal=0 then begin
      while not eof(input) do begin
         c := input^;
         output^ := c;
         put(output);
         read(input,c);
      end;
   end

   else if showoriginal=1 then begin
      while not eof(input) do begin
         if eoln(input) then writeln(output,'eoln!');
         c := input^;
         if c = '&'
         then writeln(output)
         else if c = '+'
         then write(output,' ')
         else begin
            output^ := c;
            put(output);
         end;
         read(input,c);
      end
   end

   else if showoriginal=2 then begin
      inname := true;
      inhex := false;
      while not eof(input) do begin
{
writeln(output,'inname =',inname);
writeln(output,'namelength = ',namelength:1);
}
         c := input^;

         if c = '%'
         then begin
            inhex := true;
            hexnumber := 0;
            hexcount := 0;
         end
         else if inhex then begin
{write(output,'[',c,'-',hexcount:1,']');}
            c := decapitalize(c);
{
writeln(output,'hex c = ',c:1);
writeln(output,'zero hexnumber=',hexnumber:1);
}
            if (ord(c) >= ord('0')) and (ord(c) <= ord('9')) then begin
               hexnumber := 16* hexnumber + ord(c) - ord('0');
               hexcount := succ(hexcount);
{
writeln(output,'first hexnumber=',hexnumber:1);
}
            end
            else if (ord(c) >= ord('a')) and (ord(c) <= ord('e')) then begin
               hexnumber := 16*hexnumber + ord(c) - ord('a') + 10;
{
writeln(output,'second hexnumber=',hexnumber:1);
               write(output,'"',chr(hexnumber),'"');
}
               write(output,chr(hexnumber));
               inhex := false;
            end
            else begin
               writeln(output,'Error reading hexidecimal number, character = ',c);
               halt
            end
         end

         else if inname then begin
            if c = '='
            then begin
               inname := false;
            end
            else begin
               namelength := succ(namelength);
               if namelength > maxstring
               then inname := false
               else typestring[namelength] := c
            end;

         end

         else begin (* end of variable *)
            if (c = '&') or eoln(input)
            then begin
               write(output,' ');
               for n := 1 to namelength do write(output,typestring[n]);
               writeln(output);
               namelength := 0;
               inname := true;
               inhex := false;
            end
            else if c = '+'
            then write(output,' ')
            else begin
               output^ := c;
               put(output);
            end;
         end;

         if hexcount > 1 then begin
            write(output,chr(hexnumber));
{
;write(output,'hexnumber = ',hexnumber:1);
}
            hexcount := 0;
            inhex := false;
         end;

         read(input,c);
      end
   end

end;
(* end module parseform.themain *)

begin
   themain;
{(parseformp);
}
1: end.

{
         if eoln(input) then
            writeln(output,'eoln!')
         end
         else if inname then begin
}
