program pcs(pcsp, list, output);
(* pcs: partial chi squared

  Tom Schneider
  network address: toms@alum.mit.edu


  1991
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of pcs.p 1991 August 23
origin 1991 Aug 4 *)
(* end module version *)

(* begin module describe.pcs *)
(*
name
   pcs: partial chi squared

synopsis
   pcs(pcsp: in, list: out, output: out)

files
   pcsp:  A series of lines, each of which contains
     of 4 integers, representing the numbers of a,c,g,t.
   list: partial chi squares calculated for the integers
   output: messages to the user

description
   calculate the partial chi squared values of 4 bases

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.pcs *)

(* begin module pcs.const *)
   wid = 10; (* width of display reals *)
   dec = 5; (* decimals of display reals *)
(* end module pcs.const *)

var
   list: text; (* list of results *)
   pcsp: text; (* parameter file *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module pcs.chi *)
function chi(observed, expected: real): real;
(* calculate the partial chi squared for the two inputs *)
begin
   chi := sqr(observed - expected) / expected;
end;
(* end module pcs.chi *)

(* begin module pcs.displaychi *)
procedure displaychi(var f: text; name: char; number: real;
                         partialchi, totalchi: real);
(* display the chi squared information about the named symbol to file f *)
begin
   writeln(f, '   ',name,'   ',round(number):5,
           ' ',partialchi:9:2,' ', round(100*partialchi/totalchi):5,'%');
end;
(* end module pcs.displaychi *)

(* begin module pcs.themain *)
procedure themain(var pcsp,list: text);
(* the main procedure of the program *)
var
   a,c,g,t: real; (* numbers of the 4 bases *)
   chisquared: real; (* total chi squared *)
   chia, chic, chig, chit: real; (* partial chi squared values for the bases *)
   expected: real; (* expected base composition *)
   total: real; (* sum of the 4 bases *)
begin
   writeln(output,'pcs ',version:4:2);
   rewrite(list);
   writeln(list,'pcs ',version:4:2);
   reset(pcsp);

   while not eof(pcsp) do begin
      readln(pcsp,a,c,g,t);
      total := a+c+g+t;
      expected := total/4.0;

      chia := chi(a,expected);
      chic := chi(c,expected);
      chig := chi(g,expected);
      chit := chi(t,expected);

      chisquared := chia+chic+chig+chit;

      writeln(list);
      writeln(list,'symbol  number  partial  % of');
      writeln(list,'                chi^2    total');
      displaychi(list,'a',a,chia,chisquared);
      displaychi(list,'c',c,chic,chisquared);
      displaychi(list,'g',g,chig,chisquared);
      displaychi(list,'t',t,chit,chisquared);
      writeln(list,' Total ',round(total):5,
                   ' ',chisquared:9:2,'   100%');
   end;
end;
(* end module pcs.themain *)

begin
   themain(pcsp, list);
1: end.
