program pizza(pizzap, epizza, output);
(* pizza: make an electronic pizza

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu (use only if first address fails)
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.30; (* of pizza.p 2006 Apr 18
2006 Apr 18, 1.30: syntax correction: (y/ -x) -> (y/ (-x))
1994 Sep  6, 1.29: functional
origin 1993 June 6 *)
(* end module version *)

(* begin module describe.pizza *)
(*
name
   pizza: make an electronic pizza

synopsis
   pizza(pizzap: in, epizza: out, output: out)

files
   pizzap:  paramters to control the program
      first   line:  integer, height of pizza in characters
      second  line:  integer, width of pizza in characters
      third   line:  real,    fraction covered by sauce and by cheese
      fourth  line:  integer, number of toppings to add
      fifth   line:  real,    0 < seed < 1, random number seed
      sixth   line:  real,    lowest angle of slice
      seventh line:  real,    highest angle of slice
   epizza:  you e-pizza, thermodynamically steaming hot!
   output: messages to the user

description
   The program creates an electronic pizza.

examples
   With this parameter file for pizzap:

42     height of pizza in characters
82     width of pizza in characters
0.25   fraction cover of sauce
300    number of toppings
0.3    random number seed
150    lowest angle of slice
180    highest angle of slice

we are served the epizza:

***********************************************************

             ..
           .....
         .......=]
        .....==[~==.
       .....======..=
      ....===O=~~=....~
     ....=.===={#==~====.
    .....=[.==)=======.====
   .....==.(]====~===.==(===
   ....====)==.#..=[=~&==(=.=.
  .....~.=.=======.&#======.=.==
  ....=====~@=.=$========~O===R==~
  ....~=====]==.~====.===!====.====
  ....===.H=.=.==.~.=..H=~===~==H=~~=
  ....$=.=[.====~===~RO=H==.~=====~===.
  .....=={=={==.=.========.]=.~.=$.===~.

key to this electronic pizza:
 . crust
 ~ sauce
 = cheese
 * meatball
 O olive
 & anchovy (lop off the top part and you have a little fish)
 ] [ green peppers
 } { red peppers
 =) (= mushrooms
 /\ pineapple
 $ coin (heads or tails, 8 bits = 1 byte)
 R red herring (the head is on top, tail on bottom)
 @ pepperoni
 # spinach
 H ham (aka pork)
 ! little dried hot pepper seeds

Your chef, Pizza version 1.28, says:  "Have a Byte!"

***********************************************************

documentation

see also

author
   Thomas Dana Schneider

bugs
   YUCK!

technical notes

*)
(* end module describe.pizza *)

(* begin module const.pizza *)
   tinside = 200; (* the maximum edge of the pie tin in characters *)
   shrink = 0.9; (* factor by which to shrink the pizza in the tin,
     so it fits nicely and doesn't have edges cut off *)
   toparea = 0.8; (* area of the toppings *)
(* end module const.pizza *)

type
   tin = record
      (* the pie tin in which we make the pizza *)
      height, width: integer; (* the actual size of the side *)
      surface: array[0..tinside,0..tinside] of char; (* the pizza *)
   end;

var
   pizzap, epizza: text; (* files used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module random *) 
procedure random(var seed: real); 
(* random generator 2.
   version = 1.01 of random.2 1990 Oct 2
   origin 1986 December 31
   Test this routine with the program tstrnd.
   written by David Masternarde *)

(* This random number generator is based on a shift register with a single bit
of feedback, as described in Electronics for Neurobiologists, by Brown, Maxfield
and Moraff, MIT press 1973, referencing Random Process Simulation and
Measurement by Korn, McGraw-Hill 1966.  The random seed rand, a number between 0
and 1 exclusive, is converted to an integer between 1 and 2**23-1, inclusive.
This 23-bit number is shifted right one bit and the output of the last (23rd)
bit and the 9th bit are added modulo 2 (exclusive orred) and fed back into the
new first bit.  This is done between 4 and 11 times, depending on the last 3
bits of the original number.  The result is converted back to a real number
between 0 and 1 from which the 23 bit integer can be recovered on the next call.
The 23-bit shift register goes through all 2**23-1 values before repeating;
the repetition frequency of this algorithm could be less or greater depending
on the seed, because of the random number of multiple shifts per call. *)
(* powers of 2 *)
const pow14=16384; pow15=32768; pow22= 4194304; pow23=8388608;
var iseed, (* integer shift register *)
    i, nrep: integer;  (* index, number of times to do shift *)
begin (* random *)  
      iseed := round(seed*pow23);  (* convert to 23 bit number *)
      if (iseed<1) or (iseed>=pow23) then iseed := 1;
      nrep := 4 + iseed mod 8;  (* do it 4 to 11 times based on last 3 bits *)
      for i:= 1 to nrep do
         (* if last bit and 9th bit are equal, feed back a 0, otherwise a 1 *)
	 if( odd(iseed) = ((iseed mod pow15) >= pow14))
	    then iseed := iseed div 2
	    else iseed := (iseed div 2) + pow22;
      seed := iseed/pow23;
end; (* random *) 
(* end module random *) 

(* begin module pizza.wash *)
procedure wash(var plate: tin; height,width: integer);
(* chose a tin of a given size and wash it out *)
var
   column: integer; (* the column we are printing *)
   line: integer; (* the line we are printing *)
begin
   plate.height := height;
   plate.width := width;
   for line := 1 to plate.height do begin
     for column := 1 to plate.width do begin
         plate.surface[line,column] := ' '
     end;
   end;
end;
(* end module pizza.wash *)

(* begin module pizza.ellipse *)
function ellipse(x, y, A, B, r: real): boolean;
(* is the point x, y inside the ellipse A B with "radius" r? *)
begin
  ellipse :=  sqr( (x   - (A/2)) /(A/2) )
            + sqr( (y - (B/2 )) /(B/2) )
            <= r
end;
(* end module pizza.circle *)

(* begin module pizza.adddough *)
procedure adddough(var plate: tin);
(* add doe to the pizza *)
var
   column: integer; (* the column we are printing *)
   line: integer; (* the line we are printing *)
begin
   with plate do begin
     for line := 1 to height do begin
       for column := 1 to width do begin
           if ellipse(line,column, height, width, shrink)
           then surface[line,column] := '.'
       end;
     end;
   end;
end;
(* end module pizza.adddough *)

(* begin module pizza.addsauceandcheese *)
procedure addsauceandcheese(var plate: tin; fraction: real; var seed: real);
(* add sauce and cheese to the pizza, at some fraction of coverage each.
give the random seed. *)
var
   column: integer; (* the column we are printing *)
   line: integer; (* the line we are printing *)
begin
   with plate do begin
     for line := 1 to height do begin
       for column := 1 to width do begin
           if ellipse(line,column, height, width, shrink*toparea)
           then begin
              random(seed);
              if seed <= fraction then
              surface[line,column] := '~'; (* add the sauce *)
              random(seed);
              if seed <= 1 - fraction then
              surface[line,column] := '='  (* add the cheese *)
           end;
       end;
     end;
   end;
end;
(* end module pizza.addsauceandcheese *)

(* begin module pizza.addtoppings *)
procedure addtoppings(var plate: tin;
                      toppings: integer;
                      var seed: real);
(* add a number of toppings to the plate, random generator
seed must be supplied. *)
const
   kinds = 15; (* kinds of toppings *)
var
   column: integer; (* the column we are printing *)
   line: integer; (* the line we are printing *)
   toss: integer; (* the number of the topping we are adding *)
   t: integer; (* a particular topping *)
begin
   with plate do begin
     toss := 0;
     while toss < toppings do begin
       random(seed);
       line := trunc(seed*height);
       random(seed);
       column := trunc(seed*width);

       if ellipse(line,column, height, width, shrink*toparea)
       then begin
         random(seed);
         (* hard wire the toppings for now *)
         t := round(kinds*seed);
         case t of
            0: surface[line,column] := '*';
            1: surface[line,column] := 'O';
            2: surface[line,column] := '&';
            3: surface[line,column] := '[';
            4: surface[line,column] := ']';
            5: begin
               surface[line,column]   := '=';
               surface[line,column+1] := ')';
            end;
            6: begin
               surface[line,column]   := '(';
               surface[line,column+1] := '=';
            end;
            7: begin
               surface[line,column]   := '/';
               surface[line,column+1] := '\';
            end;
            8: surface[line,column] := '$';
            9: surface[line,column] := '{';
           10: surface[line,column] := '}';
           11: surface[line,column] := 'R';
           12: surface[line,column] := '@';
           13: surface[line,column] := '#';
           14: surface[line,column] := 'H';
           15: surface[line,column] := '!';
         end;
         toss := toss + 1
       end;

     end;
   end;
end;
(* end module pizza.addtoppings *)

(* begin module pizza.key *)
procedure key(var customer: text);
(* tell the customer the key of the pizza *)
begin
writeln(customer);
writeln(customer,'key to this electronic pizza:');
writeln(customer,' . crust'); (* T. Schneider *)
writeln(customer,' ~ sauce'); (* T. Schneider *)
writeln(customer,' = cheese'); (* T. Schneider *)
writeln(customer,' * meatball'); (* T. Schneider *)
writeln(customer,' O olive'); (* T. Schneider *)
writeln(customer,' & anchovy (lop off the top part and you have a little',
                 ' fish)'); (* T. Schneider *)
writeln(customer,' ] [ green peppers'); (* Malcolm? *)
writeln(customer,' } { red peppers'); (* Malcolm? *)
writeln(customer,' =) (= mushrooms'); (* T. Schneider *)
writeln(customer,' /\ pineapple'); (* T. Schneider *)
writeln(customer,' $ coin (heads or tails, 8 bits = 1 byte)');
writeln(customer,' R red herring (the head is on top, tail on bottom)');
writeln(customer,' @ pepperoni'); (* Una Smith *)
{
already taken:
writeln(customer,' $ ham (aka pork)'); (* Una Smith *)
}
writeln(customer,' # spinach'); (* Una Smith *)
writeln(customer,' H ham (aka pork)'); (* T. Schneider *)
writeln(customer,' ! little dried hot pepper seeds'); (* Venkatesh Murthy,
                                                      venk@u.washington.edu *)
writeln(customer);

write(customer,'Your chef, Pizza version ',version:4:2);
writeln(customer,', says:  "Have a Byte!"');
writeln(customer);
end;
(* end module pizza.key *)

(* begin module toangle *)
procedure toangle(x,y: real; var angle: real);
(* convert coordinate x, y to an angle in degrees *)
const
   pi = 3.14159;
var
   d: real; (* conversion from radians to degress *)
begin
  d := 360/(2*pi);
  (* protect against division by zero, handle all zero cases *)
  if y = 0.0 then begin
            if x >= 0.0 then angle := 0
       else if x <  0.0 then angle := 180
  end
  else if x = 0.0 then begin
          if y >= 0.0 then angle := 90
     else if y <  0.0 then angle := 270
  end

  (* handle all non-zero cases *)
  else begin
         if (x > 0.0) and (y > 0.0) then angle :=       arctan( y /( x))*d
    else if (x < 0.0) and (y > 0.0) then angle := 180 - arctan( y /(-x))*d
    else if (x < 0.0) and (y < 0.0) then angle := 180 + arctan(-y /(-x))*d
    else if (x > 0.0) and (y < 0.0) then angle := 270 + arctan( x /(-y))*d
  end;

end;
(* end module toangle *)

(* begin module pizza.slice *)
procedure slice(var plate: tin; lowangle, highangle: real);
(* extract a slice of the epizza for the customer
between the low and high angles. *)
var
   angle: real; (* the angle we are printing at *)
   column: integer; (* the column we are printing *)
   h2: real; (* height / 2 to speed calculations *)
   line: integer; (* the line we are printing *)
   w2: real; (* width / 2  to speed calculations*)
   x: real; (* the x coordinate we are printing *)
   y: real; (* the y coordinate we are printing *)
begin
   with plate do begin
     h2 := height/2;
     w2 := width/2;
     for line := 1 to height do begin
       y := (height + 1 - line) - h2;
       for column := 1 to width do begin
         x := column - w2;
         toangle(x,y,angle);

         if not ((lowangle <= angle ) and (angle <= highangle))
         then plate.surface[line, column] := ' ' (* what a waste! *)
       end;
     end;
   end;
end;
(* end module pizza.slice *)

(* begin module pizza.serve *)
procedure serve(var customer: text; plate: tin);
(* serve the epizza to the customer *)
var
   column: integer; (* the column we are printing *)
   foundstart: boolean; (* we found the start of the pizza on the plate *)
   foundstop: boolean; (* we found the stop of the pizza on the plate *)
   line: integer; (* the line we are printing *)
   maximumprinting: integer; (* the last column to print.  By not
      printing spaces after the last character, we save space *)
   startline: integer; (* the line to start serving from *)
   stopline: integer; (* the line to stop serving from *)
begin
   (* find the start of the pizza *)
   startline := 0;
   foundstart := false;
   while (not foundstart) and (startline < plate.height) do begin
     startline := succ(startline);
     column := 0;
     while (column < plate.width) and (not foundstart) do begin
        column := succ(column);
        if plate.surface[startline, column]<> ' '
        then foundstart := true
     end;
   end;

   (* find the end of the pizza *)
   stopline := plate.height;
   foundstop := false;
   while (not foundstop) and (stopline > 1) do begin
     stopline := pred(stopline);
     column := 0;
     while (column < plate.width) and (not foundstop) do begin
        column := succ(column);
        if plate.surface[stopline, column]<> ' '
        then foundstop := true
     end;
   end;

   (* write out the pizza in the found region *)
   for line := startline to stopline do begin
     (* find the highest column to print in *)
     maximumprinting := plate.width;
     repeat maximumprinting := pred(maximumprinting)
     until (plate.surface[line, maximumprinting]<> ' ')
           or (maximumprinting = 1);

     for column := 1 to maximumprinting do begin
         write(customer,plate.surface[line,column])
     end;
     writeln(customer);
   end;
end;
(* end module pizza.serve *)

(* begin module pizza.themain *)
procedure themain(var pizzap, epizza: text);
(* the main procedure of the program *)
var
   highangle: real; (* lowest angle of slice *)
   lowangle: real; (* lowest angle of slice *)
   plate: tin; (* the serving dish *)
   height: integer; (* the height of the dish *)
   seed: real; (* random number generator seed *)
   sauce: real; (* fraction of sauce *)
   toppings: integer; (* number of toppings to put on *)
   width: integer; (* the width of the dish *)
begin
   writeln(output,'pizza ',version:4:2);

   reset(pizzap);
   readln(pizzap, height);
   readln(pizzap, width);
   readln(pizzap, sauce);
   readln(pizzap, toppings);
   readln(pizzap, seed);
   if (height > tinside) or (width > tinside) then begin
      writeln(output, 'requested pizza too big');
      halt
   end;
   readln(pizzap, lowangle);
   readln(pizzap, highangle);

   rewrite(epizza);

   wash(plate, height, width);
   adddough(plate);
   addsauceandcheese(plate, seed, sauce);
   addtoppings(plate, toppings, seed);
   slice(plate, lowangle, highangle);
   serve(epizza, plate);
   key(epizza);

end;
(* end module pizza.themain *)

begin
   themain(pizzap, epizza);
1: end.
