program power(input, powerp, output);
(* power: raise powerp value to the input value

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.00; (* of power.p 2012 Jun 18
2012 Jun 18, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.power *)
(*
name
   power: raise powerp value to the input value

synopsis
   power(input: in, powerp: in, output: out)

files

   input:  a series of numbers, one per line

   output:  base raised to each number

   powerp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      base: the base to use

      wid: integer width in characters of output numbers

      dec: integer decimal places of output numbers

   output: 

description

   The purpose and use of the program.

examples

   If the base is 2 (defined in powerp) with width 10 and 6 decimal
   places and the input file contains:

      1
      2
      3
      4
      5
      6

   The output will be:

  2.000000
  4.000000
  8.000000
 16.000000
 32.000000
 64.000000

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.power *)

var
   powerp (* file used by this program *)
   : text;

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module power.themain *)
procedure themain(var input, powerp, bfile: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   base: real; (* parameter version number *)
   wid: integer; (* width in characters of output numbers *)
   dec: integer; (* decimal places of output numbers *)
   exponent: real; (* an exponent read in *)
begin
{
   writeln(output,'power ',version:4:2);
}
   reset(powerp);
   readln(powerp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(powerp, base);
   readln(powerp, wid);
   readln(powerp, dec);

   if base <= 0 then begin
      writeln(output,'The base to compute a power must be positive.');
      halt;
   end;

   reset(input);
   rewrite(output);

   while not eoln(input) do begin
      readln(input, exponent);
      writeln(output, exp(exponent*ln(base)):wid:dec);
   end;

end;
(* end module power.themain *)

begin
   themain(input, powerp, output);
1: end.
