program rnclean(rnin, rncleanp, rnout, output);
(* rnclean: clean a .newsrc file

  Tom Schneider
  email: toms@alum.mit.edu


  1994
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.04; (* of rnclean.p 1994 July 2
origin 1994 June 30 *)
(* end module version *)

(* begin module describe.rnclean *)
(*
name
   rnclean: clean a .newsrc file

synopsis
   rnclean(rnin: in, rncleanp: in, rnout: out, output: out)

files
   rnin:  a .newsrc file to be cleaned
   rncleanp:  parameters to control rnclean.
     If the first line of the file contains the character 'w', the
     highest read article number will be wiped to 1.  This clears
     the newsrc file.
   rnout:  a cleaned .newsrc file
   output: messages to the user

description
   Abstract: The news reader program rn uses a file to keep track of articles
   read.  Unfortunately the lines in the file get so long eventually that the
   file cannot be read in by the vi editor.  This prevents one from turning on
   and off groups by hand.  The rnclean program cleans the .newsrc file given
   in rnin, and puts the result into rnout.

   This small tool might be widely useful, and save space.  I sometimes want to
   subscribe or unsubscribe to news groups quickly.  The news reader rn makes
   this difficult.  Instead, I edit my ~/.newsrc file, which has many lines
   that look like this:

bionet.info-theory: 1-2273

   The ":" means I am subscribed to that group.  A "!" means that I am not.
   Following this is a list of the groups that I have read and that the news
   reader has seen.  Unfortunately this list gets extremely long and after a
   while it is so long that the vi editor won't allow the lines.  Besides, it's
   a waste of space.  So I wrote a unix script (rntidy) that calls a small
   Pascal program (rnclean.p) which cleans up the list.  They are both
   available by anonymous ftp from ftp.ncifcrf.gov in pub/delila.

   The rntidy script is:

   #!/bin/csh -f
   #(ie run the cshell on this but don't read the .cshrc)

   # rntidy:  tidy up your ~/.newsrc file

   cd
   cp .newsrc .newsrc.old
   mv .newsrc /tmp/rnin
   cd /tmp
   rnclean
   mv rnout ~/.newsrc
   rm -f rnin

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.rnclean *)

var
   rnin, rncleanp, rnout: text; (* files used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module rnclean.themain *)
procedure themain(var rnin, rncleanp, rnout: text);
(* the main procedure of the program *)
var
   c: char; (* a character in rnin *)
   number: integer; (* an article number *)
   wipe: boolean; (* if true, wipe out the record of the higest
      news article read *)
begin
   writeln(output,'rnclean ',version:4:2);
   reset(rnin);
   reset(rncleanp);
   wipe := false;
   if not eof(rncleanp) then begin
      if rncleanp^='w' then wipe := true;
   end;
   if wipe
   then writeln(output,'highest news article number is now set to 1');
   rewrite(rnout);

   while not eof(rnin) do begin
      repeat
         read(rnin,c);
         write(rnout,c);
      until (c = '!') or (c = ':');

      (* build numbers *)
      number := 0;
      while not eoln(rnin) do begin
         read(rnin, c);
         if c in [' ',',','-','0','1','2','3','4','5','6','7','8','9']
         then case c of
            ' ': number := 0;
            ',': number := 0;
            '-': number := 0;
            '0': number := 10*number    ;
            '1': number := 10*number + 1;
            '2': number := 10*number + 2;
            '3': number := 10*number + 3;
            '4': number := 10*number + 4;
            '5': number := 10*number + 5;
            '6': number := 10*number + 6;
            '7': number := 10*number + 7;
            '8': number := 10*number + 8;
            '9': number := 10*number + 9;
         end
         else begin
            write  (output,'ignoring strange character found: "',c,'"');
            writeln(output,', ord(',c,') = ',ord(c):1);
         end;
      end;
      readln(rnin);

      if number < 1 then number := 1;
      if wipe then number := 1;
      writeln(rnout, ' 1-',number:1);
   end;
end;
(* end module rnclean.themain *)

begin
   themain(rnin, rncleanp, rnout);
1: end.
