program setdate(input, output);
(* setdate: set the date

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of setdate.p 1999 Dec 13
1999 Dec 13, 1.02: y2k upgrade
origin 1993 July 13 *)
(* end module version *)

(* begin module describe.setdate *)
(*
name
   setdate: set the date

synopsis
   setdate(input: in, output: out)

files
   input:  user instructions for setting the date
   output: the date for passage to the unix date function

description
   Running this program should make setting the date easier.  The program
   produces the current date in the form required by the Unix date program:

usage: date [-a sss.fff] [-u] [+format] [yymmddhhmm[.ss]]

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.setdate *)

{
   namelength = 20; (* maximum key name length *)
type
   alpha = packed array[1..namelength] of char; (* this is not alfa *)
}

(* begin module datetime.const *)
   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)
(* end module datetime.const version = 7.40; {of delmod.p 2000 Feb 18} *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.
*)
begin
   date(adatetime,'%Y/%m/%d %H:%M:%S');
end;
(* end module getdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module readdatetime  *)
procedure readdatetime (var thefile: text; var adatetime: datetimearray);
(* read the date and time from the file *)
var
      index: integer; (* to the udatetime *)
      (* the following is an unpacked date time array, to avoid reading
      into a packed array.  reading into a packed array is not transportable *)
      udatetime: array[1..datetimearraylength] of char;
begin
      for index:=1 to datetimearraylength do read(thefile,udatetime[index]);
      pack(udatetime, 1, adatetime);
      if (adatetime[3]='/') and (adatetime[12]=':') then begin
         writeln(output,' old datetime (only 2 year digits) read: ',
                          adatetime:datetimearraylength);
      end;
end;
(* end module readdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module writedatetime *)
procedure writedatetime(var thefile: text; adatetime: datetimearray);
(* expand the date and time out and print in the file *)
var
      index: integer; (* index of datetime *)
begin
      for index:=1 to datetimearraylength do write(thefile,adatetime[index])
end;
(* end module writedatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module timeseed *)
procedure addtoseed(var seed, power: real; c: char);
(* add the digit represented by c to the seed at the power position *)
var
      n: integer; (* the character represented by c *)
begin (* addtoseed *)
      power := power/10;
{
writeln(output,'addtoseed, c = ',c);
writeln(output,'addtoseed, ord(c) = ',ord(c));
}
      case c of
         ' ': begin
                 writeln(output,'timeseed: error in datetime');
                 halt;
              end;
         '0': n := 0;
         '1': n := 1;
         '2': n := 2;
         '3': n := 3;
         '4': n := 4;
         '5': n := 5;
         '6': n := 6;
         '7': n := 7;
         '8': n := 8;
         '9': n := 9
      end;
(*writeln(output,'timeseed number is [',n:1,']'); (@ debug *)
      seed := seed + power*n
end; (* addtoseed *)

procedure makeseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, reversed order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[19]);
   addtoseed(seed, power, adatetime[18]);
   (* : *)
   addtoseed(seed, power, adatetime[16]);
   addtoseed(seed, power, adatetime[15]);
   (* : *)
   addtoseed(seed, power, adatetime[13]);
   addtoseed(seed, power, adatetime[12]);
   (*   *)
   addtoseed(seed, power, adatetime[10]);
   addtoseed(seed, power, adatetime[9]);
   (* / *)
   addtoseed(seed, power, adatetime[7]);
   addtoseed(seed, power, adatetime[6]);
   (* / *)
   addtoseed(seed, power, adatetime[4]);
   addtoseed(seed, power, adatetime[3]);
   addtoseed(seed, power, adatetime[2]);
   addtoseed(seed, power, adatetime[1]);
end;

procedure orderseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, normal order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[1]);
   addtoseed(seed, power, adatetime[2]);
   addtoseed(seed, power, adatetime[3]);
   addtoseed(seed, power, adatetime[4]);
   (* / *)
   addtoseed(seed, power, adatetime[6]);
   addtoseed(seed, power, adatetime[7]);
   (* / *)
   addtoseed(seed, power, adatetime[9]);
   addtoseed(seed, power, adatetime[10]);
   (*   *)
   addtoseed(seed, power, adatetime[12]);
   addtoseed(seed, power, adatetime[13]);
   (* : *)
   addtoseed(seed, power, adatetime[15]);
   addtoseed(seed, power, adatetime[16]);
   (* : *)
   addtoseed(seed, power, adatetime[18]);
   addtoseed(seed, power, adatetime[19]);
end;

procedure timeseed(var seed: real);
(* read the computer date and time.  reverse the order of the digits
and put a decimal point in front.  this gives a fraction between
zero and one that varies quite quickly, and is always unique (if the
computer has sufficient accuracy).  it is to be used as a seed to
a random number generator. *)
var
   adatetime: datetimearray; (* a date and time *)
begin (* timeseed *)
   getdatetime(adatetime);
{
writeln(output,'timeseed: adatetime: ',adatetime);
}
   makeseed(adatetime, seed);
end; (* timeseed *)
(* end module timeseed version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module demo.time *)
procedure demotime(var fout: text);
(* write the time to file fout *)
var   dateandtime: datetimearray; (* the date and time *)
      seed: real; (* a seed for a random number generater, made
         from the date and time written backwards *)
begin (* demotime *)
   getdatetime(dateandtime);
   writeln(fout);
   writeln(fout,'The date and time is: ');
   writedatetime(fout, dateandtime);
   writeln(fout,' <- This should be the current time and date.');
   write  (fout,'1980/06/09 18:49:11');
   writeln(fout,' <- Times and dates should look like this');
   writeln(fout,'year mo da ho mi se <- with parts in these positions');
   writeln(fout);

   timeseed(seed);
   writeln(fout,'A timeseed is ',seed:14:12);
   writeln(fout,'Timeseeds can be used to start random number generators.')
end; (* demotime *)
(* end module demo.time version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module setdate.themain *)
procedure themain;
(* the main procedure of the program *)
var
   adatetime: datetimearray; (* the date and time *)
   c: char; (* a character read from the user *)
   done: boolean; (* done setting the date *)
begin
   writeln(output,'setdate ',version:4:2);
{      demotime(output); }
   getdatetime(adatetime);
   writedatetime(output,adatetime);
   writeln(output);
   done := false;

   while not done do begin
(*
      'usage: date [-a sss.fff] [-u] [+format] [yymmddhhmm[.ss]]');
*)
      writeln(output);
      write(output,'date ');

      (* year *)
      write(output, adatetime[1]);
      write(output, adatetime[2]);
      write(output, adatetime[3]);
      write(output, adatetime[4]);

      (* month *)
      write(output, adatetime[6]);
      write(output, adatetime[7]);

      (* day *)
      write(output, adatetime[9]);
      write(output, adatetime[10]);

      (* hour *)
      write(output, adatetime[12]);
      write(output, adatetime[13]);

      (* minute *)
      write(output, adatetime[15]);
      write(output, adatetime[16]);
      write(output, '.');

      (* second *)
      write(output, adatetime[18]);
      write(output, adatetime[19]);
      writeln(output);
      writeln(output);

      readln(input,c);
      if not (c in ['q','d','h','m','s'])
      then writeln(output,'commands: q(uit), d(ay),',
                          ' h(our), m(inute), s(econds)')
      else case c of
         'q': done := true;

         'd': begin
            writeln(output, 'give day dd:');
            readln(input,adatetime[7],adatetime[8]);
         end;

         'h': begin
            writeln(output, 'give hour hh:');
            readln(input,adatetime[10],adatetime[11]);
         end;

         'm': begin
            writeln(output, 'give minute mm:');
            readln(input,adatetime[13],adatetime[14]);
         end;

         's': begin
            writeln(output, 'give seconds ss:');
            readln(input,adatetime[16],adatetime[17]);
         end;

      end;
   end
end;
(* end module setdate.themain *)

begin
   themain;
1: end.
