program uncsv(csv, uncsvp, spacetable, output);
(* uncsv: convert csv to plain space table

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of uncsv.p 2011 Sep 16
2011 Sep 16, 1.01: functional
2011 Sep 15, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.uncsv *)
(*
name
   uncsv: convert csv to plain space table

synopsis
   uncsv(csv: in, uncsvp: in, spacetable: out, output: out)

files

   csv:  comma separated table

   spacetable:  space sparated table.

   uncsvp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      not used.

   output: messages to the user

description

   Convert csv table to space table.  Things inside quotes are
   preserved except spaces become underscores '_'.  Commas between
   columns become spaces.  Two commas outside quotes become a dash.

   The script unmkcsv can't handle quotes, this should.

examples

   E3017",3,"52,584,834",28,"C","G","splicing",,,"c.4348G>C","SPLICE"

becomes:

   "E3017" 3 "52,584,834" 28 "C" "G" "splicing" - - "c.4348G>C" "SPLICE"


documentation

see also

   {unmkcsv script}

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.uncsv *)

var
   csv, (* file used by this program *)
   uncsvp, (* file used by this program *)
   spacetable: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module uncsv.themain *)
procedure themain(var csv, uncsvp, spacetable: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   inquote: boolean; (* are we inside a quote? *)
   c: char; (* character transferred from csv to spacetable *)
   p: char; (* previous value of character c on this line *)
begin
   writeln(output,'uncsv ',version:4:2);
{
   reset(uncsvp);
   readln(uncsvp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

   reset(csv);
   rewrite(spacetable);
   inquote := false;

   p := ' ';
   while not eof(csv) do begin
      if eoln(csv) then begin
         readln(csv);
         writeln(spacetable);
         p := ' ';
      end
      else begin
         read(csv,c);
         if inquote then begin
            if c = '"' then begin
               inquote := not inquote;
               write(spacetable,c);
            end
            else begin
               if c = ' '
               then write(spacetable,'_')
               else write(spacetable,c);
            end;
         end
         else begin
            if c = ','
            then begin
               (* if previous character was a comma, put a dash in the
               column! *)
               if p = ',' 
               then write(spacetable,'- ')
               else write(spacetable,' ')
            end
            else write(spacetable,c);
            if c = '"' then inquote := not inquote;
         end;
         p := c;
      end;
   end;

end;
(* end module uncsv.themain *)

begin
   themain(csv, uncsvp, spacetable);
1: end.
