program vernam(input, output);
(* print the version name of a program

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

   module libraries: delman, prgmods *)

label 1; (* the end of the program *)

const
(* begin module version *)
   version = 1.00; (* of vernam.p 2011 May 06
2011 May 06, 1.00: origin from vernum
*)
(* end module version *)

(* begin module describe.vernam *)
(*
name
   vernam: print the name of a program associated with the version.

synopsis
   vernam(input: in, output: out)

files

   input: a program source code, with a version constant in the form
       "version = " followed by a real number, ' of ' and a file name.

   output: the new version number is reported.
      If there is none, the program reports 0.

description
   The program finds the version name of a file and reports it to
   output for the purpose of saving copies.

author
   Thomas Schneider

see also
   ver.p, verbop.p, vernum.p, code.p

bugs
   none known

*)
(* end module describe.vernam *)


(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const version = 'prgmod 3.96  85 mar 18 tds'; *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module vernam.themain *)
procedure themain(var source: text);
(* the themain procedure of the program *)
const
   debugging = false; (* set to true to debug *)
var
   s: char; (* a letter from the source *)
   t: trigger; (* the version trigger *)
   theversion: real; (* the version *)
begin
   (* the program cannot report itself
   writeln(output, ' vernam ', version:4:2); *)

   (*                     1         2         3         4         5 *)
   (*            12345678901234567890123456789012345678901234567890 *)
   filltrigger(t, 'version =                                         ');
   resettrigger(t);

   repeat
      testfortrigger(source^, t);
      get(source);
   until eof(source) or t.found;

   if t.found then begin
      read(source, theversion);
      {writeln(output,theversion:4:2)}
if debugging then writeln(output,theversion:4:2);
      (* skip ' of ' *)
      while source^ <> 'o' do begin
         read(source, s);
if debugging then writeln(output,'BUBBA 1 "', s,'"');
      end;
      while source^ <> 'f' do begin
         read(source, s);
if debugging then writeln(output,'BUBBA 2 "', s,'"');
      end;
      read(source, s); (* skip f *)
      read(source, s); (* skip space *)
if debugging then writeln(output,'BUBBA 3 "', s,'"');
      while source^ <> ' ' do begin
        read(source, s);
if debugging then writeln(output,'BUBBA 4 "', s,'"');
        write(output, s);
      end;
      writeln(output);
   end;
   {else writeln(output,0.0:4:2);}
end;
(* end module vernam.themain *)

begin
      themain(input);
1: end. (* vernam *)
