program winscan(winscanp, lidelp, fin, output);
(* winscan: window scan

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of winscan.p 1996 April 8
origin 1996 April 6 *)
(* end module version *)

(* begin module describe.winscan *)
(*
name
   winscan: window scan

synopsis
   winscan(winscanp: in, lidelp: out, fin: out, output: out)

files
   winscanp:  Parameters to control the program: one per line

      organism name: the entire line is used
      chromosome name: the entire line is used
      piece name: all non blank characters to the end of the line or
         the first blank character.
      start coordinate (integer)
      number of windows (integer)
      increment between windows (integer)
      window from (integer)
      window to (integer)

   lidelp: input to the lidel program created by this program

   fin:  input to the lidel program created by this program

   output: messages to the user

description
   Create delila instructions for a series of sequence windows.

examples

documentation

see also
   lidel.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.winscan *)

var
   winscanp, (* file used by this program *)
   lidelp, (* file used by this program *)
   fin: text; (* file used by this program *)

type
   nametype = record (* a name *)
      letters: array[1..50] of char;
      length: integer;
   end;

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.15; (@ of prgmod.p 1994 November 12 *)

procedure readname(var afile: text; var name: nametype);
(* Read the name from a file.  Stop at first blank. *)
begin
   name.length := 0;
   while (not eoln(afile)) and (afile^ <> ' ') do begin
      name.length := succ(name.length);
      read(afile, name.letters[name.length]);
   end;
   readln(afile);
end;

procedure writename(var afile: text; name: nametype);
(* write the name to a file *)
var
   p: integer; (* position in the name *)
begin
   for p := 1 to name.length do write(afile,name.letters[p]);
end;

(* begin module winscan.themain *)
procedure themain(var winscanp, lidelp, fin: text);
(* the main procedure of the program *)
var

   PieceName: nametype; (* name of the piece *)
   OrganismName: nametype; (* name of the Organism *)
   ChromosomeName: nametype; (* name of the Chromosome *)
   StartCoordinate: integer; (* start coordinate for windows *)
   IncrementBetweenWindows: integer; (* increment to next window zero *)
   NumberOfWindows: integer; (* number of windows *)
   w: integer; (* index to the windows *)
   WindowFrom: integer; (* delila from *)
   WindowTo: integer; (* delila to *)
   WindowCoordinate: integer; (* current coordinate of window *)

begin
   writeln(output,'winscan ',version:4:2);

   (* read parameters in *)
   reset(winscanp);
   readname(winscanp,OrganismName);
   readname(winscanp,ChromosomeName);
   readname(winscanp,PieceName);
   readln(winscanp,StartCoordinate);
   readln(winscanp,NumberOfWindows);
   readln(winscanp,IncrementBetweenWindows);
   readln(winscanp,WindowFrom);
   readln(winscanp,WindowTo);

   (* write parameters out *)
   rewrite(fin);
   writeln(fin,'* winscan ',version:4:2);
   write(fin,'* '); writename(fin, OrganismName); writeln(fin,' (organism)');
   write(fin,'* '); writename(fin, ChromosomeName); writeln(fin,' (chromosome)');
   write(fin,'* '); writename(fin, PieceName); writeln(fin,' (piece)');
   writeln(fin,'* ',StartCoordinate:1,' start coordinate');
   writeln(fin,'* ',NumberOfWindows:1,' number of windows');
   writeln(fin,'* ',IncrementBetweenWindows:1,' increment between windows');
   writeln(fin,'* ',WindowFrom:1,' window from');
   writeln(fin,'* ',WindowTo:1,' window to');

   (* create lidelp *)
   rewrite(lidelp);
   writeln(lidelp,WindowFrom:1,' ',WindowTo:1,' fromrange torange');
   writeln(lidelp,'1 column number of piece name');
   writeln(lidelp,'2 column number of coordinate data');
   writeln(lidelp,'CDS (fake value)');
   writename(lidelp, OrganismName); writeln(lidelp);
   writename(lidelp, ChromosomeName); writeln(lidelp);
   writeln(lidelp,'o   symmetry (e)ven or (o)dd, (l)eft or (r)ight');
   writeln(lidelp,'1   number of instructions to write to inst');
   writeln(lidelp,'parameters for lidel.p created by winscan ',version:4:2);


   if (WindowTo - WindowFrom + 1) > IncrementBetweenWindows
   then begin
      writeln(output,'WINDOWS OVERLAP');
      writeln(fin,'* WINDOWS OVERLAP');
   end
   else begin
      writeln(output,'windows do not overlap');
      writeln(fin,'* windows do not overlap');
   end;

   (* make the fin *)

   WindowCoordinate := StartCoordinate;
   for w := 1 to NumberOfWindows do begin
      writename(fin, PieceName);
      write(fin,' ',WindowCoordinate:1);
      writeln(fin);
      WindowCoordinate := WindowCoordinate + IncrementBetweenWindows;
   end;

end;
(* end module winscan.themain *)

begin
   themain(winscanp, lidelp, fin);
1: end.






























